@@ -72,13 +72,14 @@
 union align
 {
   /* Types which are likely to have the longest RELEVANT alignment
-   * restrictions... we don't do much with doubles.
+   * restrictions...
    */
   
   char *cp;
   void (*f)();
   long l;
   FILE *fp;
+  double d;
 };
 
 #define CLICK_SZ (sizeof(union align))
@@ -383,6 +384,16 @@ char *pstrdup(struct pool *a, const char *s)
   return res;
 }
 
+char *pstrndup(struct pool *a, const char *s, int n)
+{
+  char *res;
+  if (s == NULL) return NULL;
+  res = palloc (a, n + 1);
+  strncpy (res, s, n);
+  res[n] = '\0';
+  return res;
+}
+
 char *pstrcat(pool *a, ...)
 {
   char *cp, *argp, *res;
@@ -553,7 +564,7 @@ char *table_get (table *t, char *key)
     return NULL;
 }
 
-void table_set (table *t, char *key, char *val)
+void table_set (table *t, const char *key, const char *val)
 {
     table_entry *elts = (table_entry *)t->elts;
     int i;
@@ -569,6 +580,34 @@ void table_set (table *t, char *key, char *val)
     elts->val = pstrdup (t->pool, val);
 }
 
+void table_unset( table *t, char *key ) 
+{
+    table_entry *elts = (table_entry *)t->elts;
+    int i;   
+    int j;   
+ 
+    for (i = 0; i < t->nelts; ++i)
+        if (!strcasecmp (elts[i].key, key)) {
+ 
+            /* found the element to skip over
+             * there are any number of ways to remove an element from
+             * a contiguous block of memory.  I've chosen one that
+             * doesn't do a memcpy/bcopy/array_delete, *shrug*...
+             */
+            j = i;
+            ++i;
+            for ( ; i < t->nelts; ) {
+                elts[j].key = elts[i].key;
+                elts[j].val = elts[i].val;
+                ++i;
+                ++j;
+            };
+            --t->nelts;
+
+            return;
+        }
+}     
+
 void table_merge (table *t, char *key, char *val)
 {
     table_entry *elts = (table_entry *)t->elts;
@@ -585,6 +624,15 @@ void table_merge (table *t, char *key, char *val)
     elts->val = pstrdup (t->pool, val);
 }
 
+void table_add (table *t, char *key, char *val)
+{
+    table_entry *elts = (table_entry *)t->elts;
+
+    elts = (table_entry *)push_array(t);
+    elts->key = pstrdup (t->pool, key);
+    elts->val = pstrdup (t->pool, val);
+}
+
 table* overlay_tables (pool *p, table *overlay, table *base)
 {
     return append_arrays (p, overlay, base);
@@ -841,14 +889,24 @@ int spawn_child (pool *p, void (*func)(void *), void *data,
   
   if (pipe_out) {
     close (out_fds[1]);
+#ifdef __EMX__
+    /* Need binary mode set for OS/2. */
+    *pipe_out = fdopen (out_fds[0], "rb");
+#else
     *pipe_out = fdopen (out_fds[0], "r");
-    
+#endif  
+  
     if (*pipe_out) note_cleanups_for_file (p, *pipe_out);
   }
 
   if (pipe_in) {
     close (in_fds[0]);
+#ifdef __EMX__
+    /* Need binary mode set for OS/2 */
+    *pipe_in = fdopen (in_fds[1], "wb");
+#else
     *pipe_in = fdopen (in_fds[1], "w");
+#endif
     
     if (*pipe_in) note_cleanups_for_file (p, *pipe_in);
   }
@@ -914,3 +972,99 @@ static void free_proc_chain (struct process_chain *procs)
       waitpid (p->pid, &status, 0);
   }
 }
+
+#ifdef __EMX__
+int spawn_child_os2 (pool *p, void (*func)(void *), void *data,
+         enum kill_conditions kill_how,
+         FILE **pipe_in, FILE **pipe_out, char *buffer, int lenp)
+{
+  int pid;
+  int in_fds[2];
+  int out_fds[2];
+
+  block_alarms();
+  
+  if (pipe_in && pipe (in_fds) < 0)
+  {
+      unblock_alarms();
+      return 0;
+  }
+  
+  if (pipe_out && pipe (out_fds) < 0) {
+    if (pipe_in) {
+      close (in_fds[0]); close (in_fds[1]);
+    }
+    unblock_alarms();
+    return 0;
+  }
+
+  if ((pid = fork()) < 0) {
+    if (pipe_in) {
+      close (in_fds[0]); close (in_fds[1]);
+    }
+    if (pipe_out) {
+      close (out_fds[0]); close (out_fds[1]);
+    }
+    unblock_alarms();
+    return 0;
+  }
+
+  if (!pid) {
+    int stdinpipe[2];
+    /* Due to a limitation of EMX, inheriting socket handles is not
+    allowed so we need to read the input and place it in a pipe and 
+    then pass that handle instead of the socket. */
+    
+    if (lenp > 0) {
+        pipe(stdinpipe);
+        write(stdinpipe[1], buffer, lenp);
+        close(stdinpipe[1]);
+        in_fds[0] = dup(stdinpipe[0]);
+        close(stdinpipe[0]);
+    }
+                                
+    /* Child process */
+    
+    if (pipe_out) {
+      close (out_fds[0]);
+      dup2 (out_fds[1], STDOUT_FILENO);
+      close (out_fds[1]);
+    }
+
+    if (pipe_in) {
+      close (in_fds[1]);
+      dup2 (in_fds[0], STDIN_FILENO);
+      close (in_fds[0]);
+    }
+
+    /* HP-UX SIGCHLD fix goes here, if someone will remind me what it is... */
+    signal (SIGCHLD, SIG_DFL);    /* Was that it? */
+    
+    func (data);
+    exit (0);            /* Should never get here... */
+  }
+
+  /* Parent process */
+
+  note_subprocess (p, pid, kill_how);
+  
+  if (pipe_out) {
+    close (out_fds[1]);
+    /* Need binary mode set for OS/2. */
+    *pipe_out = fdopen (out_fds[0], "rb");
+    
+    if (*pipe_out) note_cleanups_for_file (p, *pipe_out);
+  }
+
+  if (pipe_in) {
+    close (in_fds[0]);
+    /* Need binary mode set for OS/2. */
+    *pipe_in = fdopen (in_fds[1], "wb");
+    
+    if (*pipe_in) note_cleanups_for_file (p, *pipe_in);
+  }
+
+  unblock_alarms();
+  return pid;
+}
+#endif
