@@ -54,6 +54,7 @@
 
 /*
  * conf.h: system-dependant #defines and includes...
+ * See README for a listing of what they mean
  */
 
 #ifndef QNX
@@ -68,8 +69,9 @@
 #undef NO_SETSID
 char *crypt(char *pw, char *salt);
 #define JMP_BUF sigjmp_buf
-#define NEED_STRERROR
+#define HAVE_MMAP
 #include <sys/time.h>     
+#define NEED_STRERROR
 
 #elif defined(SOLARIS2)
 #undef HAS_GMTOFF
@@ -80,6 +82,8 @@ char *crypt(char *pw, char *salt);
 #define getwd(d) getcwd(d,MAX_STRING_LEN)
 #define JMP_BUF sigjmp_buf
 #define FCNTL_SERIALIZED_ACCEPT
+#define HAVE_MMAP
+#define HAVE_CRYPT_H
 
 #elif defined(IRIX)
 #undef HAS_GMTOFF
@@ -87,6 +91,8 @@ char *crypt(char *pw, char *salt);
 #undef NO_SETSID
 #define JMP_BUF sigjmp_buf
 #define FCNTL_SERIALIZED_ACCEPT
+#define HAVE_SHMGET
+#define HAVE_CRYPT_H
  
 #elif defined(HPUX)
 #define HAVE_RESOURCE 1
@@ -98,6 +104,7 @@ char *crypt(char *pw, char *salt);
 #endif
 #define getwd(d) getcwd(d,MAX_STRING_LEN)
 #define JMP_BUF sigjmp_buf
+#define HAVE_MMAP
 
 #elif defined(AIX)
 #undef HAS_GMTOFF
@@ -123,6 +130,8 @@ char *crypt(char *pw, char *salt);
 #undef NO_KILLPG
 #undef NO_SETSID
 #define JMP_BUF sigjmp_buf
+#define HAVE_MMAP
+#define HAVE_CRYPT_H
 
 #elif defined(SEQUENT)
 #define HAS_GMTOFF
@@ -175,6 +184,7 @@ typedef unsigned short mode_t;
 #define waitpid(a,b,c) wait4(a,(union wait *)b,c,NULL)
 typedef int pid_t;
 #define JMP_BUF jmp_buf
+#define NO_USE_SIGACTION
 
 #elif defined(LINUX)
 #undef HAS_GMTOFF
@@ -183,6 +193,7 @@ typedef int pid_t;
 #undef NEED_STRDUP
 #define JMP_BUF sigjmp_buf
 #define FCNTL_SERIALIZED_ACCEPT
+#include <sys/time.h>     
 
 #elif defined(SCO)
 #undef HAS_GMTOFF
@@ -197,6 +208,15 @@ typedef int pid_t;
 
 #define JMP_BUF sigjmp_buf
 #define SIGURG SIGUSR1
+#define NEED_SELECT_H
+#define FCNTL_SERIALIZED_ACCEPT
+#define HAVE_MMAP
+#define SecureWare
+
+/* Although SCO 5 defines these in <strings.h> (note the "s") they don't have
+consts. Sigh. */
+extern int strcasecmp(const char *,const char *);
+extern int strncasecmp(const char *,const char *,unsigned);
 
 #elif defined(CONVEXOS)
 #define HAS_GMTOFF
@@ -209,6 +229,13 @@ typedef int pid_t;
 #undef NO_SETSID
 #define NEED_STRDUP
 #define JMP_BUF sigjmp_buf
+/* fcntl() locking is expensive with NFS */
+#undef FLOCK_SERIALIZED_ACCEPT
+#define HAVE_SHMGET
+#define MOVEBREAK		0x4000000
+/* These are to let -Wall compile more cleanly */
+extern int strcasecmp(const char *, const char *);
+extern int strncasecmp(const char *,const char *,unsigned);
 
 #elif defined(SVR4)
 #define NO_KILLPG
@@ -222,6 +249,18 @@ typedef int pid_t;
 /* A lot of SVR4 systems need this */
 #define FCNTL_SERIALIZED_ACCEPT
 
+#elif defined(DGUX)
+#define NO_KILLPG
+#undef  NO_SETSID
+#undef NEED_STRDUP
+#define NEED_STRCASECMP
+#define NEED_STRNCASECMP
+#define bzero(a,b) memset(a,0,b)
+#define JMP_BUF sigjmp_buf
+#define getwd(d) getcwd(d,MAX_STRING_LEN)
+/* A lot of SVR4 systems need this */
+#define FCNTL_SERIALIZED_ACCEPT
+
 #elif defined(__NetBSD__)
 #define HAS_GMTOFF
 #undef NO_KILLPG
@@ -250,11 +289,13 @@ typedef int pid_t;
 #undef NO_KILLPG
 #undef NO_SETSID
 #define JMP_BUF sigjmp_buf
+#define HAVE_MMAP
 
 #elif defined(QNX)
 #undef NO_KILLPG
 #undef NO_SETSID
 #define NEED_INITGROUPS
+#define NEED_SELECT_H
 #define JMP_BUF sigjmp_buf
 
 #elif defined(LYNXOS)
@@ -266,6 +307,16 @@ typedef int pid_t;
 #define NEED_INITGROUPS
 #define JMP_BUF jmp_buf
 
+#elif defined(__EMX__)
+/* Defines required for EMX OS/2 port. */
+#define JMP_BUF sigjmp_buf
+#define NO_KILLPG
+#define NEED_STRCASECMP
+#define NEED_STRNCASECMP
+#define NO_SETSID
+/* Add some drive name support */
+#define chdir _chdir2
+
 /* Unknown system - Edit these to match */
 #else
 #ifdef BSD
@@ -341,6 +392,16 @@ int getrlimit( int, struct rlimit *);
 int setrlimit( int, struct rlimit *);
 #endif
 #endif
+#ifdef HAVE_MMAP
+#include <sys/mman.h>
+#endif
+#if !defined(MAP_ANON) && defined(MAP_ANONYMOUS)
+#define MAP_ANON MAP_ANONYMOUS
+#endif
+
+#if defined(HAVE_MMAP) && defined(NO_MMAP)
+#undef HAVE_MMAP
+#endif
 
 #ifndef LOGNAME_MAX
 #define LOGNAME_MAX 25
