@@ -51,6 +51,12 @@
  *
  */
 
+#ifdef WWW_USER
+#include <sys/param.h>
+#include <sys/types.h>
+#include <pwd.h>
+#include <string.h>
+#endif
 
 #define CORE_PRIVATE
 #include "httpd.h"
@@ -61,6 +67,9 @@
 
 #include "http_main.h"		/* For the default_handler below... */
 #include "http_log.h"
+#include "rfc1413.h"
+#include "util_md5.h"
+#include "scoreboard.h"
 
 /* Server core module... This module provides support for really basic
  * server operations, including options and commands which control the
@@ -86,6 +95,10 @@ void *create_core_dir_config (pool *a, char *dir)
     conf->opts = dir ? OPT_UNSET : OPT_ALL;
     conf->override = dir ? OR_UNSET : OR_ALL;
 
+    conf->content_md5 = 2;
+
+    conf->hostname_lookups = 2;/* binary, but will use 2 as an "unset = on" */
+    conf->do_rfc1413 = DEFAULT_RFC1413 | 2;  /* set bit 1 to indicate default */
     return (void *)conf;
 }
 
@@ -112,6 +125,10 @@ void *merge_core_dir_configs (pool *a, void *basev, void *newv)
     for (i = 0; i <= RESPONSE_CODES; ++i)
         if (new->response_code_strings[i] != NULL)
 	   conf->response_code_strings[i] = new->response_code_strings[i];
+    if (new->hostname_lookups != 2)
+	conf->hostname_lookups = new->hostname_lookups;
+    if ((new->do_rfc1413 & 2) == 0) conf->do_rfc1413 = new->do_rfc1413;
+    if ((new->content_md5 & 2) == 0) conf->content_md5 = new->content_md5;
 
     return (void*)conf;
 }
@@ -120,11 +137,40 @@ void *create_core_server_config (pool *a, server_rec *s)
 {
     core_server_config *conf =
       (core_server_config *)pcalloc(a, sizeof(core_server_config));
-    int is_virtual = is_virtual_server (s);
+    int is_virtual = s->is_virtual;
   
     conf->access_name = is_virtual ? NULL : DEFAULT_ACCESS_FNAME;
-    conf->document_root = is_virtual ? NULL : DOCUMENT_LOCATION;
+
+    /*
+     * The root of document is either the home of the WWW_USER or
+     * what was defined in the configuration.
+     *
+     */
+
+    if (is_virtual) {
+	conf->document_root = 0;
+    } else {
+#ifdef WWW_USER
+	struct passwd* www_user = getpwnam(WWW_USER);
+	if (www_user && *www_user->pw_dir) {
+	    static char www_htdocs[MAXPATHLEN + 1];
+
+	    strcpy(www_htdocs, www_user->pw_dir);
+#ifdef WWW_USER_DOCUMENT_LOCATION
+	    strcat(strcat(www_htdocs, "/"), WWW_USER_DOCUMENT_LOCATION);
+#endif
+
+	    conf->document_root = www_htdocs;
+	} else {
+	    conf->document_root = DOCUMENT_LOCATION;
+	}
+#else
+	conf->document_root = DOCUMENT_LOCATION;
+#endif
+    }
+
     conf->sec = make_array (a, 40, sizeof(void *));
+    conf->sec_url = make_array (a, 40, sizeof(void *));
     
     return (void *)conf;
 }
@@ -140,6 +186,7 @@ void *merge_core_server_configs (pool *p, void *basev, void *virtv)
     if (!conf->access_name) conf->access_name = base->access_name;
     if (!conf->document_root) conf->document_root = base->document_root;
     conf->sec = append_arrays (p, virt->sec, base->sec);
+    conf->sec_url = append_arrays (p, virt->sec_url, base->sec_url);
 
     return conf;
 }
@@ -157,6 +204,15 @@ void add_per_dir_conf (server_rec *s, void *dir_config)
     *new_space = dir_config;
 }
 
+void add_per_url_conf (server_rec *s, void *url_config)
+{
+    core_server_config *sconf = get_module_config (s->module_config,
+						   &core_module);
+    void **new_space = (void **) push_array (sconf->sec_url);
+    
+    *new_space = url_config;
+}
+
 /*****************************************************************
  *
  * There are some elements of the core config structures in which
@@ -236,6 +292,87 @@ char *response_code_string (request_rec *r, int error_index)
     return conf->response_code_strings[error_index];
 }
 
+const char *
+get_remote_host(conn_rec *conn, void *dir_config, int type)
+{
+    struct in_addr *iaddr;
+    struct hostent *hptr;
+#ifdef MAXIMUM_DNS
+    char **haddr;
+#endif
+    core_dir_config *dir_conf;
+
+/* If we haven't checked the host name, and we want to */
+    dir_conf = (core_dir_config *)get_module_config(dir_config, &core_module);
+
+    if (conn->remote_host == NULL && dir_conf->hostname_lookups)
+    {
+#ifdef STATUS
+	int old_stat = update_child_status(conn->child_num,
+						SERVER_BUSY_DNS,
+						(request_rec*)NULL);
+#endif /* STATUS */
+	iaddr = &(conn->remote_addr.sin_addr);
+	hptr = gethostbyaddr((char *)iaddr, sizeof(struct in_addr), AF_INET);
+	if (hptr != NULL)
+	{
+	    conn->remote_host = pstrdup(conn->pool, (void *)hptr->h_name);
+	    str_tolower (conn->remote_host);
+	   
+#ifdef MAXIMUM_DNS
+    /* Grrr. Check THAT name to make sure it's really the name of the addr. */
+    /* Code from Harald Hanche-Olsen <hanche@imf.unit.no> */
+
+	    hptr = gethostbyname(conn->remote_host);
+	    if (hptr)
+	    {
+		for(haddr=hptr->h_addr_list; *haddr; haddr++)
+		    if(((struct in_addr *)(*haddr))->s_addr == iaddr->s_addr)
+			break;
+	    }
+	    if((!hptr) || (!(*haddr)))
+		conn->remote_host = NULL;
+#endif
+	}
+/* if failed, set it to the NULL string to indicate error */
+	if (conn->remote_host == NULL) conn->remote_host = "";
+#ifdef STATUS
+	(void)update_child_status(conn->child_num,old_stat,(request_rec*)NULL);
+#endif /* STATUS */
+    }
+
+/*
+ * Return the desired information; either the remote DNS name, if found,
+ * or either NULL (if the hostname was requested) or the IP address
+ * (if any identifier was requested).
+ */
+    if (conn->remote_host != NULL && conn->remote_host[0] != '\0')
+	return conn->remote_host;
+    else
+    {
+	if (type == REMOTE_HOST) return NULL;
+	else return conn->remote_ip;
+    }
+}
+
+const char *
+get_remote_logname(request_rec *r)
+{
+    core_dir_config *dir_conf;
+
+    if (r->connection->remote_logname != NULL)
+	return r->connection->remote_logname;
+
+/* If we haven't checked the identity, and we want to */
+    dir_conf = (core_dir_config *)
+	get_module_config(r->per_dir_config, &core_module);
+
+    if (dir_conf->do_rfc1413 & 1)
+	return rfc1413(r->connection, r->server);
+    else
+	return NULL;
+}
+
 /*****************************************************************
  *
  * Commands... this module handles almost all of the NCSA httpd.conf
@@ -256,7 +393,11 @@ char *set_document_root (cmd_parms *cmd, void *dummy, char *arg)
     void *sconf = cmd->server->module_config;
     core_server_config *conf = get_module_config (sconf, &core_module);
   
-    if (!is_directory (arg)) return "DocumentRoot must be a directory";
+    if (!is_directory (arg))
+	if (cmd->server->is_virtual)
+	    fprintf (stderr, "Warning: DocumentRoot [%s] does not exist\n", arg);
+	else
+	    return "DocumentRoot must be a directory";
     
     conf->document_root = arg;
     return NULL;
@@ -384,6 +525,7 @@ char *limit (cmd_parms *cmd, void *dummy, char *arg)
 	else if(!strcasecmp(method,"PUT")) limited |= (1 << M_PUT);
 	else if(!strcasecmp(method,"POST")) limited |= (1 << M_POST);
 	else if(!strcasecmp(method,"DELETE")) limited |= (1 << M_DELETE);
+        else if(!strcasecmp(method,"CONNECT")) limited |= (1 << M_CONNECT);
 	else return "unknown method in <Limit>";
     }
 
@@ -430,6 +572,43 @@ char *dirsection (cmd_parms *cmd, void *dummy, char *arg)
     return errmsg;
 }
 
+static char *end_url_magic = "</Location> outside of any <Location> section";
+
+char *end_urlsection (cmd_parms *cmd, void *dummy) {
+    return end_url_magic;
+}
+
+char *urlsection (cmd_parms *cmd, void *dummy, char *arg)
+{
+    char *errmsg, *endp = strrchr (arg, '>');
+    int old_overrides = cmd->override;
+    char *old_path = cmd->path;
+    core_dir_config *conf;
+
+    void *new_url_conf = create_per_dir_config (cmd->pool);
+
+    if (endp) *endp = '\0';
+
+    if (cmd->path) return "<Location> sections don't nest";
+    if (cmd->limited != -1) return "Can't have <Location> within <Limit>";
+    
+    cmd->path = getword_conf (cmd->pool, &arg);
+    cmd->override = OR_ALL|ACCESS_CONF;
+
+    errmsg = srm_command_loop (cmd, new_url_conf);
+    if (errmsg != end_url_magic) return errmsg;
+
+    conf = (core_dir_config *)get_module_config(new_url_conf, &core_module);
+    conf->d = pstrdup(cmd->pool, cmd->path);	/* No mangling, please */
+
+    add_per_url_conf (cmd->server, new_url_conf);
+    
+    cmd->path = old_path;
+    cmd->override = old_overrides;
+
+    return NULL;
+}
+
 /* httpd.conf commands... beginning with the <VirtualHost> business */
 
 char *end_virthost_magic = "</Virtualhost> out of place";
@@ -446,7 +625,7 @@ char *virtualhost_section (cmd_parms *cmd, void *dummy, char *arg)
 
     if (endp) *endp = '\0';
     
-    if (is_virtual_server (main_server))
+    if (main_server->is_virtual)
 	return "<VirtualHost> doesn't nest!";
     
     s = init_virtual_host (p, arg);
@@ -514,6 +693,16 @@ char *set_timeout (cmd_parms *cmd, void *dummy, char *arg) {
     return NULL;
 }
 
+char *set_keep_alive_timeout (cmd_parms *cmd, void *dummy, char *arg) {
+    cmd->server->keep_alive_timeout = atoi (arg);
+    return NULL;
+}
+
+char *set_keep_alive (cmd_parms *cmd, void *dummy, char *arg) {
+    cmd->server->keep_alive = atoi (arg);
+    return NULL;
+}
+
 char *set_pidfile (cmd_parms *cmd, void *dummy, char *arg) {
     pid_fname = pstrdup (cmd->pool, arg);
     return NULL;
@@ -524,8 +713,24 @@ char *set_scoreboard (cmd_parms *cmd, void *dummy, char *arg) {
     return NULL;
 }
 
-char *set_idcheck (cmd_parms *cmd, void *dummy, int arg) {
-    cmd->server->do_rfc931 = arg;
+char *set_idcheck (cmd_parms *cmd, core_dir_config *d, int arg) {
+    d->do_rfc1413 = arg;
+    return NULL;
+}
+
+char *set_hostname_lookups (cmd_parms *cmd, core_dir_config *d, int arg) {
+    d->hostname_lookups = arg;
+    return NULL;
+}
+
+char *set_serverpath (cmd_parms *cmd, void *dummy, char *arg) {
+    cmd->server->path = pstrdup (cmd->pool, arg);
+    cmd->server->pathlen = strlen (arg);
+    return NULL;
+}
+
+char *set_content_md5 (cmd_parms *cmd, core_dir_config *d, int arg) {
+    d->content_md5 = arg;
     return NULL;
 }
 
@@ -553,6 +758,8 @@ char *set_max_free_servers (cmd_parms *cmd, void *dummy, char *arg) {
 
 char *set_server_limit (cmd_parms *cmd, void *dummy, char *arg) {
     daemons_limit = atoi (arg);
+    if (daemons_limit > HARD_SERVER_LIMIT)
+    	daemons_limit = HARD_SERVER_LIMIT;
     return NULL;
 }
 
@@ -562,13 +769,40 @@ char *set_max_requests (cmd_parms *cmd, void *dummy, char *arg) {
 }
 
 char *set_bind_address (cmd_parms *cmd, void *dummy, char *arg) {
-    bind_address.s_addr = get_virthost_addr (arg, 1);
+    bind_address.s_addr = get_virthost_addr (arg, NULL);
     return NULL;
 }
 
-/* Note --- change the mask below, and ErrorDocument will work from
- * .htaccess files.  The question is, what AllowOverride should webmasters
- * have to turn it off?
+char *set_listener(cmd_parms *cmd, void *dummy, char *ips)
+{
+    listen_rec *new;
+    char *ports;
+
+    if (cmd->server->is_virtual) return "Listen not allowed in <VirtualHost>";
+    ports=strchr(ips, ':');
+    if (ports != NULL)
+    {
+	if (ports == ips) return "Missing IP address";
+	else if (ports[0] == '\0')
+	    return "Address must end in :<port-number>";
+	*(ports++) = '\0';
+    } else
+	ports = ips;
+
+    new=palloc(cmd->pool, sizeof(listen_rec));
+    new->local_addr.sin_family = AF_INET;
+    if (ports == ips) /* no address */
+	new->local_addr.sin_addr.s_addr = htonl(INADDR_ANY);
+    else
+	new->local_addr.sin_addr.s_addr = get_virthost_addr(ips, NULL);
+    new->local_addr.sin_port = htons(atoi(ports));
+    new->next = listeners;
+    listeners = new;
+    return NULL;
+}
+
+/* Note --- ErrorDocument will now work from .htaccess files.  
+ * The AllowOverride of Fileinfo allows webmasters to turn it off
  */
 
 command_rec core_cmds[] = {
@@ -577,6 +811,8 @@ command_rec core_cmds[] = {
 
 { "<Directory", dirsection, NULL, RSRC_CONF, RAW_ARGS, NULL },
 { "</Directory>", end_dirsection, NULL, ACCESS_CONF, NO_ARGS, NULL },
+{ "<Location", urlsection, NULL, RSRC_CONF, RAW_ARGS, NULL },
+{ "</Location>", end_urlsection, NULL, ACCESS_CONF, NO_ARGS, NULL },
 { "<Limit", limit, NULL, OR_ALL, RAW_ARGS, NULL },
 { "</Limit>", endlimit, NULL, OR_ALL, RAW_ARGS, NULL },
 { "AuthType", set_string_slot, (void*)XtOffsetOf(core_dir_config, auth_type),
@@ -589,7 +825,7 @@ command_rec core_cmds[] = {
   
 { "AccessFileName", set_access_name, NULL, RSRC_CONF, TAKE1, NULL },
 { "DocumentRoot", set_document_root, NULL, RSRC_CONF, TAKE1, NULL },
-{ "ErrorDocument", set_error_document, NULL, RSRC_CONF, RAW_ARGS, NULL },
+{ "ErrorDocument", set_error_document, NULL, OR_FILEINFO, RAW_ARGS, NULL },
 { "AllowOverride", set_override, NULL, ACCESS_CONF, RAW_ARGS, NULL },
 { "Options", set_options, NULL, OR_OPTIONS, RAW_ARGS, NULL },
 { "DefaultType", set_string_slot,
@@ -600,6 +836,7 @@ command_rec core_cmds[] = {
 
 { "ServerType", server_type, NULL, RSRC_CONF, TAKE1,"'inetd' or 'standalone'"},
 { "Port", server_port, NULL, RSRC_CONF, TAKE1, "a TCP port number"},
+{ "HostnameLookups", set_hostname_lookups, NULL, ACCESS_CONF|RSRC_CONF, FLAG, NULL },
 { "User", set_user, NULL, RSRC_CONF, TAKE1, "a username"},
 { "Group", set_group, NULL, RSRC_CONF, TAKE1, "a group name"},
 { "ServerAdmin", set_server_string_slot,
@@ -622,8 +859,16 @@ command_rec core_cmds[] = {
 { "ResourceConfig", set_server_string_slot,
   (void *)XtOffsetOf (server_rec, srm_confname), RSRC_CONF, TAKE1,
   "the filename of the resource config file" },
+{ "ServerAlias", set_server_string_slot,
+   (void *)XtOffsetOf (server_rec, names), RSRC_CONF, RAW_ARGS,
+   "a name or names alternately used to access the server" },
+{ "ServerPath", set_serverpath, NULL, RSRC_CONF, TAKE1,
+  "The pathname the server can be reached at" },
 { "Timeout", set_timeout, NULL, RSRC_CONF, TAKE1, "timeout duration (sec)"},
-{ "IdentityCheck", set_idcheck, NULL, RSRC_CONF, FLAG, NULL },
+{ "KeepAliveTimeout", set_keep_alive_timeout, NULL, RSRC_CONF, TAKE1, "Keep-Alive timeout duration (sec)"},
+{ "KeepAlive", set_keep_alive, NULL, RSRC_CONF, TAKE1, "Maximum Keep-Alive requests per connection (0 to disable)" },
+{ "IdentityCheck", set_idcheck, NULL, RSRC_CONF|ACCESS_CONF, FLAG, NULL },
+{ "ContentDigest", set_content_md5, NULL, RSRC_CONF|ACCESS_CONF|OR_AUTHCFG, FLAG, "whether or not to send a Content-MD5 header with each request" },
 { "CacheNegotiatedDocs", },
 { "StartServers", set_daemons_to_start, NULL, RSRC_CONF, TAKE1, NULL },
 { "MinSpareServers", set_min_free_servers, NULL, RSRC_CONF, TAKE1, NULL },
@@ -634,6 +879,8 @@ command_rec core_cmds[] = {
 { "MaxRequestsPerChild", set_max_requests, NULL, RSRC_CONF, TAKE1, NULL },
 { "BindAddress", set_bind_address, NULL, RSRC_CONF, TAKE1,
   "'*', a numeric IP address, or the name of a host with a unique IP address"},
+{ "Listen", set_listener, NULL, RSRC_CONF, TAKE1,
+      "a port number or a numeric IP address and a port number"},
 { "<VirtualHost", virtualhost_section, NULL, RSRC_CONF, RAW_ARGS, NULL },
 { "</VirtualHost>", end_virtualhost_section, NULL, RSRC_CONF, NO_ARGS, NULL },
 { NULL },
@@ -649,9 +896,16 @@ int core_translate (request_rec *r)
     void *sconf = r->server->module_config;
     core_server_config *conf = get_module_config (sconf, &core_module);
   
+    if (r->proxyreq) return NOT_IMPLEMENTED;
     if (r->uri[0] != '/') return BAD_REQUEST;
     
-    r->filename = pstrcat (r->pool, conf->document_root, r->uri, NULL);
+    if (r->server->path &&
+	!strncmp(r->uri, r->server->path, r->server->pathlen))
+      r->filename = pstrcat (r->pool, conf->document_root,
+			     (r->uri + r->server->pathlen), NULL);
+    else
+      r->filename = pstrcat (r->pool, conf->document_root, r->uri, NULL);
+
     return OK;
 }
 
@@ -666,10 +920,13 @@ int do_nothing (request_rec *r) { return OK; }
 
 int default_handler (request_rec *r)
 {
+    core_dir_config *d =
+      (core_dir_config *)get_module_config(r->per_dir_config, &core_module);
     int errstatus;
     FILE *f;
     
     if (r->method_number != M_GET) return DECLINED;
+
     if (r->finfo.st_mode == 0 || (r->path_info && *r->path_info)) {
 	log_reason("File does not exist", r->filename, r);
 	return NOT_FOUND;
@@ -679,13 +936,22 @@ int default_handler (request_rec *r)
 	|| (errstatus = set_last_modified (r, r->finfo.st_mtime)))
         return errstatus;
     
+#ifdef __EMX__
+    /* Need binary mode for OS/2 */
+    f = fopen (r->filename, "rb");
+#else
     f = fopen (r->filename, "r");
+#endif
 
     if (f == NULL) {
         log_reason("file permissions deny server access", r->filename, r);
         return FORBIDDEN;
     }
-      
+
+    if (d->content_md5 & 1) {
+      table_set (r->headers_out, "Content-MD5", md5digest(r->pool, f));
+    }
+
     soft_timeout ("send", r);
     
     send_http_header (r);
