@@ -62,19 +62,46 @@
 
 #include "httpd.h"
 #include "http_config.h"
+#include "http_core.h"
 #include "http_log.h"
 
 #include <stdarg.h>
 
+void error_log_child (void *cmd)
+{
+    /* Child process code for 'ErrorLog "|..."';
+     * may want a common framework for this, since I expect it will
+     * be common for other foo-loggers to want this sort of thing...
+     */
+    
+    cleanup_for_exec();
+    signal (SIGHUP, SIG_IGN);
+    execl (SHELL_PATH, SHELL_PATH, "-c", (char *)cmd, NULL);
+    exit (1);
+}
+
 void open_error_log(server_rec *s, pool *p)
 {
     char *fname;
   
     fname = server_root_relative (p, s->error_fname);
-    if(!(s->error_log = pfopen(p, fname, "a"))) {
-        fprintf(stderr,"httpd: could not open error log file %s.\n", fname);
-        perror("fopen");
-        exit(1);
+
+    if (*fname == '|') {
+      FILE *dummy;
+
+      spawn_child(p, error_log_child, (void *)(fname+1),
+                    kill_after_timeout, &dummy, NULL);
+
+        if (dummy == NULL) {
+            fprintf (stderr, "Couldn't fork child for ErrorLog process\n");
+            exit (1);
+      }
+    } else {
+        if(!(s->error_log = pfopen(p, fname, "a"))) {
+            fprintf(stderr,"httpd: could not open error log file %s.\n", fname);
+            perror("fopen");
+            exit(1);
+      }
     }
 }
 
@@ -152,12 +179,15 @@ log_printf(const server_rec *s, const char *fmt, ...)
     va_end (args);
 
     fputc('\n', s->error_log);
+    fflush(s->error_log);
 }
 
 void log_reason(char *reason, char *file, request_rec *r) {
     fprintf (r->server->error_log,
 	     "[%s] access to %s failed for %s, reason: %s\n",
-	     get_time(), file, r->connection->remote_name, reason);
+	     get_time(), file,
+	     get_remote_host(r->connection, r->per_dir_config, REMOTE_NAME),
+	     reason);
     fflush (r->server->error_log);
 }
 
