@@ -85,8 +85,25 @@
 #include "http_protocol.h"	/* for read_request */
 #include "http_request.h"	/* for process_request */
 #include "http_conf_globals.h"
+#include "http_core.h"          /* for get_remote_host */
 #include "scoreboard.h"
 #include <setjmp.h>
+#ifdef HAVE_SHMGET
+#include <sys/types.h>
+#include <sys/ipc.h>
+#include <sys/shm.h>
+#if defined(AUX)		/* Aren't defined anyplace */
+extern char *shmat(int, char *, int);
+extern int  shmctl(int, int, struct shmid_ds *);
+extern int  shmget(key_t, int, int);
+extern char *sbrk(int);
+#endif
+#endif
+#ifdef SecureWare
+#include <sys/security.h>
+#include <sys/audit.h>
+#include <prot.h>
+#endif
 
 /*
  * Actual definitions of config globals... here because this is
@@ -102,10 +119,12 @@ char *pid_fname;
 char *scoreboard_fname;
 char *server_argv0;
 struct in_addr bind_address;
+listen_rec *listeners;
 int daemons_to_start;
 int daemons_min_free;
 int daemons_max_free;
 int daemons_limit;
+time_t restart_time;
 
 char server_root[MAX_STRING_LEN];
 char server_confname[MAX_STRING_LEN];
@@ -116,6 +135,8 @@ server_rec *server_conf;
 JMP_BUF jmpbuffer;
 JMP_BUF restart_buffer;
 int sd;
+static fd_set listenfds;
+static int listenmaxfd;
 pid_t pgrp;
 
 /* one_process --- debugging mode variable; can be set from the command line
@@ -131,14 +152,14 @@ pid_t pgrp;
 
 int one_process = 0;
 
-#ifdef FCNTL_SERIALIZED_ACCEPT
+#if defined(FCNTL_SERIALIZED_ACCEPT)
 static struct flock lock_it = { F_WRLCK, 0, 0, 0 };
 static struct flock unlock_it = { F_UNLCK, 0, 0, 0 };
 
 static int lock_fd=-1;
 
 /*
- * Initialise mutex lock.
+ * Initialize mutex lock.
  * Must be safe to call this on a restart.
  */
 void
@@ -158,7 +179,7 @@ accept_mutex_init(pool *p)
     if (lock_fd == -1)
     {
 	perror ("open");
-	fprintf (stderr, "Cannot open lcok file\n");
+	fprintf (stderr, "Cannot open lock file: %s\n", lock_fname);
 	exit (1);
     }
     unlink(lock_fname);
@@ -187,6 +208,60 @@ void accept_mutex_off()
 	exit(1);
     }
 }
+#elif defined(FLOCK_SERIALIZED_ACCEPT)
+
+static int lock_fd=-1;
+
+/*
+ * Initialize mutex lock.
+ * Must be safe to call this on a restart.
+ */
+void
+accept_mutex_init(pool *p)
+{
+    char lock_fname[30];
+
+    strcpy(lock_fname, "/usr/tmp/htlock.XXXXXX");
+    
+    if (mktemp(lock_fname) == NULL || lock_fname[0] == '\0')
+    {
+	fprintf (stderr, "Cannot assign name to lock file!\n");
+	exit (1);
+    }
+
+    lock_fd = popenf(p, lock_fname, O_CREAT | O_WRONLY, 0644);
+    if (lock_fd == -1)
+    {
+	perror ("open");
+	fprintf (stderr, "Cannot open lock file\n");
+	exit (1);
+    }
+    unlink(lock_fname);
+}
+
+void accept_mutex_on()
+{
+    int ret;
+    
+    while ((ret = flock(lock_fd, LOCK_EX)) < 0 && errno == EINTR)
+	continue;
+
+    if (ret < 0) {
+	log_unixerr("flock", "LOCK_EX", "Error getting accept lock. Exiting!",
+		    server_conf);
+	exit(1);
+    }
+}
+
+void accept_mutex_off()
+{
+    if (flock (lock_fd, LOCK_UN) < 0)
+    {
+	log_unixerr("flock", "LOCK_UN", "Error freeing accept lock. Exiting!",
+		    server_conf);
+	exit(1);
+    }
+}
 #else
 /* Default --- no serialization.  Other methods *could* go here,
  * as #elifs...
@@ -223,7 +298,9 @@ void timeout(sig)			/* Also called on SIGPIPE */
 int sig;
 {
     char errstr[MAX_STRING_LEN];
+    void *dirconf;
 
+    signal(SIGPIPE, SIG_IGN);		/* Block SIGPIPE */
     if (alarms_blocked) {
 	alarm_pending = 1;
 	return;
@@ -237,17 +314,20 @@ int sig;
 #endif
     }
     
+    if (timeout_req != NULL) dirconf = timeout_req->per_dir_config;
+    else dirconf = current_conn->server->lookup_defaults;
     if (sig == SIGPIPE) {
         sprintf(errstr,"%s lost connection to client %s",
 	    timeout_name ? timeout_name : "request",
-	    current_conn->remote_name);
+	    get_remote_host(current_conn, dirconf, REMOTE_NAME));
     } else {
         sprintf(errstr,"%s timed out for %s",
 	    timeout_name ? timeout_name : "request",
-	    current_conn->remote_name);
+	    get_remote_host(current_conn, dirconf, REMOTE_NAME));
     }
     
-    log_error(errstr, current_conn->server);
+    if (!current_conn->keptalive) 
+       log_error(errstr, current_conn->server);
       
     if (timeout_req) {
 	/* Someone has asked for this transaction to just be aborted
@@ -262,12 +342,10 @@ int sig;
 	    else log_req = log_req->prev;
 	}
 	
-	log_transaction(log_req);
+	if (!current_conn->keptalive) 
+            log_transaction(log_req);
 
-	pfclose (timeout_req->connection->pool,
-		 timeout_req->connection->client);
-	pfclose (timeout_req->connection->pool,
-	         timeout_req->connection->request_in);
+	bclose(timeout_req->connection->client);
     
 	if (!standalone) exit(0);
 #ifdef NEXT
@@ -305,7 +383,10 @@ void hard_timeout (char *name, request_rec *r)
     timeout_name = name;
     
     signal(SIGALRM,(void (*)())timeout);
-    alarm (r->server->timeout);
+    if (r->connection->keptalive) 
+       alarm (r->server->keep_alive_timeout);
+    else
+       alarm (r->server->timeout);
 }
 
 void soft_timeout (char *name, request_rec *r)
@@ -345,7 +426,138 @@ void reset_timeout (request_rec *r) {
  * We begin with routines which deal with the file itself... 
  */
 
-static short_score scoreboard_image[HARD_SERVER_MAX];
+#if defined(HAVE_MMAP)
+static short_score *scoreboard_image=NULL;
+
+static void setup_shared_mem(void)
+{
+    caddr_t m;
+#if defined(MAP_ANON) || defined(MAP_FILE)
+/* BSD style */
+    m = mmap((caddr_t)0, HARD_SERVER_LIMIT*sizeof(short_score),
+	     PROT_READ | PROT_WRITE, MAP_ANON | MAP_SHARED, -1, 0);
+    if (m == (caddr_t)-1)
+    {
+	perror("mmap");
+	fprintf(stderr, "httpd: Could not mmap memory\n");
+	exit(1);
+    }
+#else
+/* Sun style */
+    int fd;
+
+    fd = open("/dev/zero", O_RDWR);
+    if (fd == -1)
+    {
+	perror("open");
+	fprintf(stderr, "httpd: Could not open /dev/zero\n");
+	exit(1);
+    }
+    m = mmap((caddr_t)0, HARD_SERVER_LIMIT*sizeof(short_score),
+	     PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
+    if (m == (caddr_t)-1)
+    {
+	perror("mmap");
+	fprintf(stderr, "httpd: Could not mmap /dev/zero\n");
+	exit(1);
+    }
+    close(fd);
+#endif
+    scoreboard_image = (short_score *)m;
+}
+
+#elif defined(HAVE_SHMGET)
+static short_score *scoreboard_image=NULL;
+static key_t shmkey = IPC_PRIVATE;
+static int shmid = -1;
+
+static void setup_shared_mem(void)
+{
+    int score_size = HARD_SERVER_LIMIT*sizeof(short_score);
+    char errstr[MAX_STRING_LEN];
+    struct shmid_ds shmbuf;
+#ifdef MOVEBREAK
+    char *obrk;
+#endif
+
+    if ((shmid = shmget(shmkey, score_size, IPC_CREAT|SHM_R|SHM_W)) == -1)
+    {
+	perror("shmget");
+	fprintf(stderr, "httpd: Could not call shmget\n");
+	exit(1);
+    }
+
+    sprintf(errstr, "created shared memory segment #%d", shmid);
+    log_error(errstr, server_conf);
+
+#ifdef MOVEBREAK
+    /*
+     * Some SysV systems place the shared segment WAY too close
+     * to the dynamic memory break point (sbrk(0)). This severely
+     * limits the use of malloc/sbrk in the program since sbrk will
+     * refuse to move past that point.
+     *
+     * To get around this, we move the break point "way up there",
+     * attach the segment and then move break back down. Ugly
+     */
+    if ((obrk=sbrk(MOVEBREAK)) == (char *)-1)
+    {
+	perror("sbrk");
+	fprintf(stderr, "httpd: Could not move break\n");
+    }
+#endif
+
+#define BADSHMAT	((short_score*)(-1))
+    if ((scoreboard_image = (short_score*)shmat(shmid, 0, 0)) == BADSHMAT)
+    {
+	perror("shmat");
+	fprintf(stderr, "httpd: Could not call shmat\n");
+	/*
+	 * We exit below, after we try to remove the segment
+	 */
+    }
+    else	/* only worry about permissions if we attached the segment */
+    {
+	if (shmctl(shmid, IPC_STAT, &shmbuf) != 0) {
+	    perror("shmctl");
+	    fprintf(stderr, "httpd: Could not stat segment #%d\n", shmid);
+	}
+	else
+	{
+	    shmbuf.shm_perm.uid = user_id;
+	    shmbuf.shm_perm.gid = group_id;
+	    if (shmctl(shmid, IPC_SET, &shmbuf) != 0) {
+		perror("shmctl");
+		fprintf(stderr, "httpd: Could not set segment #%d\n", shmid);
+	    }
+	}
+    }
+    /*
+     * We must avoid leaving segments in the kernel's
+     * (small) tables.
+     */
+    if (shmctl(shmid, IPC_RMID, NULL) != 0) {
+	perror("shmctl");
+	fprintf(stderr, "httpd: Could not delete segment #%d\n", shmid);
+	sprintf(errstr, "could not remove shared memory segment #%d", shmid);
+	log_unixerr("shmctl","IPC_RMID",errstr, server_conf);
+    }
+    if (scoreboard_image == BADSHMAT)	/* now bailout */
+	exit(1);
+
+#ifdef MOVEBREAK
+    if (obrk == (char *)-1)
+	return;		/* nothing else to do */
+    if (sbrk(-(MOVEBREAK)) == (char *)-1)
+    {
+	perror("sbrk");
+	fprintf(stderr, "httpd: Could not move break back\n");
+    }
+#endif
+}
+
+#else
+static short_score scoreboard_image[HARD_SERVER_LIMIT];
 static int have_scoreboard_fname = 0;
 static int scoreboard_fd;
 
@@ -378,14 +590,28 @@ static int force_read (int fd, char *buffer, int bufsz)
     
     return rv < 0? rv : orig_sz - bufsz;
 }
+#endif
 
+/* Called by parent process */
 void reinit_scoreboard (pool *p)
 {
+#if defined(HAVE_SHMGET) || defined(HAVE_MMAP)
+    if (scoreboard_image == NULL)
+    {
+	setup_shared_mem();
+    }
+    memset(scoreboard_image, 0, HARD_SERVER_LIMIT*sizeof(short_score));
+#else
     scoreboard_fname = server_root_relative (p, scoreboard_fname);
 
     have_scoreboard_fname = 1;
     
+#ifdef __EMX__
+    /* OS/2 needs binary mode set. */
+    scoreboard_fd = popenf(p, scoreboard_fname, O_CREAT|O_BINARY|O_RDWR, 0644);
+#else
     scoreboard_fd = popenf(p, scoreboard_fname, O_CREAT|O_RDWR, 0644);
+#endif
     if (scoreboard_fd == -1)
     {
 	fprintf (stderr, "Cannot open scoreboard file:\n");
@@ -396,24 +622,35 @@ void reinit_scoreboard (pool *p)
     memset ((char*)scoreboard_image, 0, sizeof(scoreboard_image));
     force_write (scoreboard_fd, (char*)scoreboard_image,
 		 sizeof(scoreboard_image));
+#endif
 }
 
+/* called by child */
 void reopen_scoreboard (pool *p)
 {
+#if !defined(HAVE_MMAP) && !defined(HAVE_SHMGET)
     if (scoreboard_fd != -1) pclosef (p, scoreboard_fd);
     
+#ifdef __EMX__    
+    /* OS/2 needs binary mode set. */
+    scoreboard_fd = popenf(p, scoreboard_fname, O_CREAT|O_BINARY|O_RDWR, 0666);
+#else
     scoreboard_fd = popenf(p, scoreboard_fname, O_CREAT|O_RDWR, 0666);
+#endif
     if (scoreboard_fd == -1)
     {
 	fprintf (stderr, "Cannot open scoreboard file:\n");
 	perror (scoreboard_fname);
 	exit (1);
     }
+#endif
 }
 
 void cleanup_scoreboard ()
 {
+#if !defined(HAVE_MMAP) && !defined(HAVE_SHMGET)
     unlink (scoreboard_fname);
+#endif
 }
 
 /* Routines called to deal with the scoreboard image
@@ -429,27 +666,126 @@ void cleanup_scoreboard ()
 
 void sync_scoreboard_image ()
 {
+#if !defined(HAVE_MMAP) && !defined(HAVE_SHMGET)
     lseek (scoreboard_fd, 0L, 0);
     force_read (scoreboard_fd, (char*)scoreboard_image,
 		sizeof(scoreboard_image));
+#endif
 }
 
-void update_child_status (int child_num, int status)
+int update_child_status (int child_num, int status, request_rec *r)
 {
     short_score new_score_rec;
+    int old_status;
+
+    if (child_num < 0)
+	return -1;
+    
+    memcpy(&new_score_rec,&scoreboard_image[child_num],sizeof new_score_rec);
     new_score_rec.pid = getpid();
+    old_status = new_score_rec.status;
     new_score_rec.status = status;
 
+#if defined(STATUS)
+    new_score_rec.last_used=time(NULL);
+    if (status == SERVER_DEAD) {
+	/*
+	 * Reset individual counters
+	 */
+	new_score_rec.my_access_count = 0L;
+	new_score_rec.my_bytes_served = 0L;
+	new_score_rec.conn_count = (unsigned short)0;
+	new_score_rec.conn_bytes = (unsigned long)0;
+    }
+    if (r) {
+	int slot_size;
+	conn_rec *c = r->connection;
+	slot_size = sizeof(new_score_rec.client) - 1;
+	strncpy(new_score_rec.client, get_remote_host(c, r->per_dir_config,
+	 REMOTE_NAME), slot_size);
+	new_score_rec.client[slot_size] = '\0';
+	slot_size = sizeof(new_score_rec.request) - 1;
+	strncpy(new_score_rec.request, (r->the_request ? r->the_request :
+	 "NULL"), slot_size);
+	new_score_rec.request[slot_size] = '\0';
+    }
+#endif
+
+#if defined(HAVE_MMAP) || defined(HAVE_SHMGET)
+    memcpy(&scoreboard_image[child_num], &new_score_rec, sizeof(short_score));
+#else
     lseek (scoreboard_fd, (long)child_num * sizeof(short_score), 0);
     force_write (scoreboard_fd, (char*)&new_score_rec, sizeof(short_score));
+#endif
+
+    return old_status;
 }
 
+int get_child_status (int child_num)
+{
+    if (child_num<0 || child_num>=HARD_SERVER_LIMIT)
+    	return -1;
+    else
+	return scoreboard_image[child_num].status;
+}
+
+int count_busy_servers ()
+{
+    int i;
+    int res = 0;
+
+    for (i = 0; i < HARD_SERVER_LIMIT; ++i)
+      if (scoreboard_image[i].status == SERVER_BUSY_READ ||
+              scoreboard_image[i].status == SERVER_BUSY_WRITE ||
+              scoreboard_image[i].status == SERVER_BUSY_KEEPALIVE ||
+              scoreboard_image[i].status == SERVER_BUSY_LOG ||
+              scoreboard_image[i].status == SERVER_BUSY_DNS)
+          ++res;
+    return res;
+}
+
+short_score get_scoreboard_info(int i)
+{
+    return (scoreboard_image[i]);
+}
+
+#if defined(STATUS)
+void increment_counts (int child_num, request_rec *r, int flag)
+{
+    long int bs=0;
+    short_score new_score_rec=scoreboard_image[child_num];
+
+    if (r->sent_bodyct)
+        bgetopt(r->connection->client, BO_BYTECT, &bs);
+
+    if (flag) {
+	new_score_rec.conn_count = (unsigned short)0;
+	new_score_rec.conn_bytes = (unsigned long)0;
+    }
+    new_score_rec.access_count ++;
+    new_score_rec.my_access_count ++;
+    new_score_rec.conn_count ++;
+    new_score_rec.bytes_served += (unsigned long)bs;
+    new_score_rec.my_bytes_served += (unsigned long)bs;
+    new_score_rec.conn_bytes += (unsigned long)bs;
+
+    times(&new_score_rec.times);
+
+#if defined(HAVE_MMAP) || defined(HAVE_SHMGET)
+    memcpy(&scoreboard_image[child_num], &new_score_rec, sizeof(short_score));
+#else
+    lseek (scoreboard_fd, (long)child_num * sizeof(short_score), 0);
+    force_write (scoreboard_fd, (char*)&new_score_rec, sizeof(short_score));
+#endif
+}
+#endif
+
 int count_idle_servers ()
 {
     int i;
     int res = 0;
 
-    for (i = 0; i < HARD_SERVER_MAX; ++i)
+    for (i = 0; i < HARD_SERVER_LIMIT; ++i)
 	if (scoreboard_image[i].status == SERVER_READY
 	  || scoreboard_image[i].status == SERVER_STARTING)
 	    ++res;
@@ -461,7 +797,7 @@ int find_free_child_num ()
 {
     int i;
 
-    for (i = 0; i < HARD_SERVER_MAX; ++i)
+    for (i = 0; i < HARD_SERVER_LIMIT; ++i)
 	if (scoreboard_image[i].status == SERVER_DEAD)
 	    return i;
 
@@ -472,7 +808,7 @@ int find_child_by_pid (int pid)
 {
     int i;
 
-    for (i = 0; i < HARD_SERVER_MAX; ++i)
+    for (i = 0; i < HARD_SERVER_LIMIT; ++i)
 	if (scoreboard_image[i].pid == pid)
 	    return i;
 
@@ -485,7 +821,7 @@ void reclaim_child_processes ()
     int my_pid = getpid();
 
     sync_scoreboard_image();
-    for (i = 0; i < HARD_SERVER_MAX; ++i) {
+    for (i = 0; i < HARD_SERVER_LIMIT; ++i) {
 	int pid = scoreboard_image[i].pid;
 
 	if (pid != my_pid && pid != 0)
@@ -497,20 +833,27 @@ void reclaim_child_processes ()
  * a while...
  */
 
-static jmp_buf wait_timeout_buf;
+static JMP_BUF wait_timeout_buf;
 static int wait_or_timeout_retval = -1;
 
 static void longjmp_out_of_alarm (int sig) {
+#ifdef NEXT
     longjmp (wait_timeout_buf, 1);
+#else
+    siglongjmp (wait_timeout_buf, 1);
+#endif
 }
 
 int wait_or_timeout (int *status)
 {
     wait_or_timeout_retval = -1;
     
+#if defined(NEXT)
     if (setjmp(wait_timeout_buf) != 0) {
+#else 
+    if (sigsetjmp(wait_timeout_buf, 1) != 0) {
+#endif
 	errno = ETIMEDOUT;
-	alarm(0);
 	return wait_or_timeout_retval;
     }
     
@@ -555,6 +898,10 @@ void detach()
         perror("setpgrp");
         exit(1);
     }
+#else
+#ifdef __EMX__
+    /* OS/2 doesn't support process group IDs */
+    pgrp=getpid();
 #else
     if((pgrp=setpgrp(getpid(),0)) == -1) {
         fprintf(stderr,"httpd: setpgrp failed\n");
@@ -563,6 +910,7 @@ void detach()
     }
 #endif    
 #endif
+#endif
 }
 
 void sig_term() {
@@ -621,13 +969,16 @@ static void set_group_privs()
       uid_t uid=atoi(&user_name[1]);
 
       if ((ent = getpwuid(uid)) == NULL) {
-	 log_error("couldn't determine user name from uid", server_conf);
+	 log_unixerr("getpwuid",NULL,"couldn't determine user name from uid", server_conf);
 	 exit(1);
       }
       
       name = ent->pw_name;
     } else name = user_name;
 
+#ifndef __EMX__ 
+    /* OS/2 dosen't support groups. */
+
     /* Reset `groups' attributes. */
     
     if (initgroups(name, group_id) == -1) {
@@ -639,6 +990,7 @@ static void set_group_privs()
 	log_unixerr("setgid", NULL, "unable to set group id", server_conf);
 	exit (1);
     }
+#endif 
   }
 }
 
@@ -653,13 +1005,26 @@ void restart() {
 }
 
 void set_signals() {
-	if(!one_process)
-	{
-	    signal(SIGSEGV,(void (*)())seg_fault);
+#ifndef NO_USE_SIGACTION
+    struct sigaction sa;
+#endif
+    if(!one_process) {
+	signal(SIGSEGV,(void (*)())seg_fault);
     	signal(SIGBUS,(void (*)())bus_error);
 	}
+
+#ifdef NO_USE_SIGACTION
     signal(SIGTERM,(void (*)())sig_term);
     signal(SIGHUP,(void (*)())restart);
+#else
+    memset(&sa,0,sizeof sa);
+    sa.sa_handler=(void (*)())sig_term;
+    if(sigaction(SIGTERM,&sa,NULL) < 0)
+	log_unixerr("sigaction(SIGTERM)", NULL, NULL, server_conf);
+    sa.sa_handler=(void (*)())restart;
+    if(sigaction(SIGHUP,&sa,NULL) < 0)
+	log_unixerr("sigaction(SIGHUP)", NULL, NULL, server_conf);
+#endif
 }
 
 /*****************************************************************
@@ -670,12 +1035,16 @@ void set_signals() {
 pool *pconf;			/* Pool for config stuff */
 pool *ptrans;			/* Pool for per-transaction stuff */
 
-server_rec *find_virtual_server (struct in_addr server_ip, server_rec *server)
+server_rec *find_virtual_server (struct in_addr server_ip, int port,
+				 server_rec *server)
 {
     server_rec *virt;
 
     for (virt = server->next; virt; virt = virt->next)
-	if (virt->host_addr.s_addr == server_ip.s_addr)
+	if ((virt->is_virtual == 1) &&	/* VirtualHost */
+	    (virt->host_addr.s_addr == htonl(INADDR_ANY) ||
+	     virt->host_addr.s_addr == server_ip.s_addr) &&
+	    (virt->host_port == 0 || virt->host_port == port))
 	    return virt;
 
     return server;
@@ -683,34 +1052,56 @@ server_rec *find_virtual_server (struct in_addr server_ip, server_rec *server)
 
 void default_server_hostnames(server_rec *s)
 {
+    struct hostent *h, *main;
+    char *def_hostname;
+    int n;
+
     /* Main host first */
     
     if (!s->server_hostname)
 	s->server_hostname = get_local_host(pconf);
 
+    def_hostname = s->server_hostname;
+    main = gethostbyname(def_hostname);
+
     /* Then virtual hosts */
     
-    for (s = s->next; s; s = s->next)
-	if (!s->server_hostname) {
-	    struct hostent *h = gethostbyaddr ((char *)&(s->host_addr),
-					       sizeof (struct in_addr),
-					       AF_INET);
-	    if (h != NULL) {
-		s->server_hostname = pstrdup (pconf, (char *)h->h_name);
+    for (s = s->next; s; s = s->next) {
+	/* Check to see if we might be a HTTP/1.1 virtual host - same IP */
+	for (n = 0; main->h_addr_list[n] != NULL; n++) {
+	  if (s->host_addr.s_addr ==
+	      (((struct in_addr *)(main->h_addr_list[n]))->s_addr))
+	    s->is_virtual = 2;
+	}
+
+        if (!s->server_hostname) {
+	    if (s->is_virtual == 2)
+	        s->server_hostname = s->virthost;
+	    else if (s->host_addr.s_addr == htonl(INADDR_ANY))
+		s->server_hostname = def_hostname;
+	    else
+	    {
+		h = gethostbyaddr ((char *)&(s->host_addr),
+				   sizeof (struct in_addr), AF_INET);
+		if (h != NULL)
+		    s->server_hostname = pstrdup (pconf, (char *)h->h_name);
 	    }
 	}
+    }
 }
 	
 void abort_connection (conn_rec *c)
 {
     /* Make sure further I/O DOES NOT HAPPEN */
-    shutdown (fileno (c->client), 2);
+    shutdown (c->client->fd, 2);
     signal (SIGPIPE, SIG_IGN);	/* Ignore further complaints */
     c->aborted = 1;
 }
 
-conn_rec *new_connection (pool *p, server_rec *server, FILE *in, FILE *out,
-			  const struct sockaddr_in *saddr)
+conn_rec *new_connection (pool *p, server_rec *server, BUFF *inout,
+			  const struct sockaddr_in *remaddr,
+			  const struct sockaddr_in *saddr,
+			  int child_num)
 {
     conn_rec *conn = (conn_rec *)pcalloc (p, sizeof(conn_rec));
     
@@ -719,14 +1110,18 @@ conn_rec *new_connection (pool *p, server_rec *server, FILE *in, FILE *out,
      */
     
     conn = (conn_rec *)pcalloc(p, sizeof(conn_rec));
+    conn->child_num = child_num;
     
     conn->pool = p;
-    conn->server = find_virtual_server (saddr->sin_addr, server);
-    conn->client = out;
-    conn->request_in = in;
-    
-    get_remote_host(conn);
+    conn->local_addr = *saddr;
+    conn->server = find_virtual_server(saddr->sin_addr, ntohs(saddr->sin_port),
+				       server);
+    conn->client = inout;
     
+    conn->remote_addr = *remaddr;
+    conn->remote_ip = pstrdup (conn->pool,
+			       inet_ntoa(conn->remote_addr.sin_addr));
+
     return conn;
 }
 
@@ -747,18 +1142,12 @@ void child_main(int child_num_arg)
     struct sockaddr sa_server;
     struct sockaddr sa_client;
 
-#ifdef ULTRIX_BRAIN_DEATH
-    extern char *rfc931();
-#else
-    extern char *rfc931 (struct sockaddr_in *, struct sockaddr_in *);
-#endif
-
     csd = -1;
     dupped_csd = -1;
     child_num = child_num_arg;
     requests_this_child = 0;
     reopen_scoreboard (pconf);
-    update_child_status (child_num, SERVER_READY);
+    (void)update_child_status (child_num, SERVER_READY, (request_rec*)NULL);
 
     /* Only try to switch if we're running as root */
     if(!geteuid() && setuid(user_id) == -1) {
@@ -771,10 +1160,12 @@ void child_main(int child_num_arg)
 #else
     sigsetjmp(jmpbuffer,1);
 #endif
+#ifndef __EMX__
     signal(SIGURG, timeout);
+#endif    
 
     while (1) {
-	FILE *conn_in, *conn_out;
+	BUFF *conn_io;
 	request_rec *r;
       
         alarm(0);		/* Cancel any outstanding alarms. */
@@ -794,13 +1185,38 @@ void child_main(int child_num_arg)
 	}
 
 	clen=sizeof(sa_client);
-	update_child_status (child_num, SERVER_READY);
+	(void)update_child_status (child_num, SERVER_READY, (request_rec*)NULL);
 	
 	accept_mutex_on();  /* Lock around "accept", if necessary */
-	
-	while ((csd=accept(sd, &sa_client, &clen)) == -1) 
-           if (errno != EINTR) 
+
+	if (listeners != NULL)
+	{
+	    fd_set fds;
+
+	    for (;;) {
+		memcpy(&fds, &listenfds, sizeof(fd_set));
+#ifdef HPUX
+		csd = select(listenmaxfd+1, (int*)&fds, NULL, NULL, NULL);
+#else
+                csd = select(listenmaxfd+1, &fds, NULL, NULL, NULL);
+#endif
+		if (csd == -1 && errno != EINTR)
+		    log_unixerr("select",NULL,"select error", server_conf);
+		if (csd <= 0) continue;
+		for (sd=listenmaxfd; sd >= 0; sd--)
+		    if (FD_ISSET(sd, &fds)) break;
+		if (sd < 0) continue;
+
+		clen=sizeof(sa_client);
+		do csd=accept(sd, &sa_client, &clen);
+		while (csd == -1 && errno == EINTR);
+		if (csd != -1) break;
 		log_unixerr("accept", "(client socket)", NULL, server_conf);
+	    }
+	} else
+	    while ((csd=accept(sd, &sa_client, &clen)) == -1) 
+		if (errno != EINTR) 
+		    log_unixerr("accept",NULL,"socket error: accept failed", server_conf);
 
 	accept_mutex_off(); /* unlock after "accept" */
 
@@ -810,36 +1226,49 @@ void child_main(int child_num_arg)
 	    continue;
 	}
 	
+	(void)update_child_status (child_num, SERVER_BUSY_READ, (request_rec*)NULL);
+	conn_io = bcreate(ptrans, B_RDWR);
 	dupped_csd = csd;
-#if defined(AUX) || defined(SCO)
+#if defined(NEED_DUPPED_CSD)
 	if ((dupped_csd = dup(csd)) < 0) {
 	    log_unixerr("dup", NULL, "couldn't duplicate csd", server_conf);
-	    dupped_csd = csd;	/* Oh well... */
+	    dupped_csd = csd;   /* Oh well... */
 	}
 #endif
-	update_child_status (child_num, SERVER_BUSY);
-	conn_in = pfdopen (ptrans, csd, "r");
-	conn_out = pfdopen (ptrans, dupped_csd, "w");
+	bpushfd(conn_io, csd, dupped_csd);
 
-	current_conn = new_connection (ptrans, server_conf, conn_in, conn_out,
-				       (struct sockaddr_in *)&sa_server);
-
-	if (current_conn->server->do_rfc931)
-	    current_conn->remote_logname = 
-		rfc931((struct sockaddr_in *)&sa_client,
-		       (struct sockaddr_in *)&sa_server);
+	current_conn = new_connection (ptrans, server_conf, conn_io,
+				       (struct sockaddr_in *)&sa_client,
+				       (struct sockaddr_in *)&sa_server,
+				       child_num);
 	
 	r = read_request (current_conn);
+	(void)update_child_status (child_num, SERVER_BUSY_WRITE, r);
 	if (r) process_request (r); /* else premature EOF --- ignore */
-		
+
+#if defined(STATUS)
+        if (r) increment_counts(child_num,r,1);
+#endif
+	while (r && current_conn->keepalive) {
+	  bflush(conn_io);
+	  destroy_pool(r->pool);
+	  (void)update_child_status (child_num, SERVER_BUSY_KEEPALIVE, (request_rec*)NULL);
+	  r = read_request (current_conn);
+	  (void)update_child_status (child_num, SERVER_BUSY_WRITE, r);
+	  if (r) process_request (r);
+
+#if defined(STATUS)
+	  if (r) increment_counts(child_num,r,0);
+#endif
+	}
+#if 0	
 	if (bytes_in_pool (ptrans) > 80000)
 	    log_printf(r->server,
 		       "Memory hog alert: allocated %ld bytes for %s",
 		       bytes_in_pool (ptrans), r->the_request);
-		
-	fflush (conn_out);
-	pfclose (ptrans,conn_in);
-	pfclose (ptrans,conn_out);
+#endif		
+	bflush(conn_io);
+	bclose(conn_io);
     }    
 }
 
@@ -865,12 +1294,69 @@ void make_child(server_rec *server_conf, int child_num)
     }
 }
 
+static int
+make_sock(pool *pconf, const struct sockaddr_in *server)
+{
+    int s;
+    const int one = 1;
+    const int keepalive_value = 1;  
+
+    if ((s = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP)) == -1) {
+        perror("socket");
+        fprintf(stderr,"httpd: could not get socket\n");
+        exit(1);
+    }
+
+    note_cleanups_for_fd (pconf, s); /* arrange to close on exec or restart */
+    
+    if((setsockopt(s, SOL_SOCKET,SO_REUSEADDR,(char *)&one,sizeof(one)))
+       == -1) {
+	perror("setsockopt(SO_REUSEADDR)");
+	fprintf(stderr,"httpd: could not set socket option SO_REUSEADDR\n");
+        exit(1);
+    }
+    if((setsockopt(s, SOL_SOCKET,SO_KEEPALIVE,(char *)&keepalive_value,
+        sizeof(keepalive_value))) == -1) {
+	perror("setsockopt(SO_KEEPALIVE)"); 
+        fprintf(stderr,"httpd: could not set socket option SO_KEEPALIVE\n"); 
+        exit(1); 
+    }
+
+#ifdef NEED_LINGER   /* If puts don't complete, you could try this. */
+    {
+	struct linger li;
+	li.l_onoff = 1;
+	li.l_linger = 900;
+
+	if (setsockopt(s, SOL_SOCKET, SO_LINGER,
+	    (char *)&li, sizeof(struct linger)) < 0) {
+	    perror("setsockopt(SO_LINGER)");
+	    fprintf(stderr,"httpd: could not set socket option SO_LINGER\n");
+	    exit(1);
+	}
+    }
+#endif  /* NEED_LINGER */
+
+    if(bind(s, (struct sockaddr *)server,sizeof(struct sockaddr_in)) == -1)
+    {
+        perror("bind");
+	if (server->sin_addr.s_addr != htonl(INADDR_ANY))
+	    fprintf(stderr,"httpd: could not bind to address %s port %d\n",
+		    inet_ntoa(server->sin_addr), ntohs(server->sin_port));
+	else
+	    fprintf(stderr,"httpd: could not bind to port %d\n",
+		    ntohs(server->sin_port));
+        exit(1);
+    }
+    listen(s, 512);
+    return s;
+}
+
+
 /*****************************************************************
  * Executive routines.
  */
 
-static int keepalive_value = 1;  
-static int one = 1;
 static int num_children = 0;
 
 void standalone_main(int argc, char **argv)
@@ -878,7 +1364,7 @@ void standalone_main(int argc, char **argv)
     struct sockaddr_in sa_server;
 
     standalone = 1;
-    sd = -1;
+    sd = listenmaxfd = -1;
     
     if (!one_process) detach(); 
     
@@ -891,17 +1377,21 @@ void standalone_main(int argc, char **argv)
     signal (SIGHUP, SIG_IGN);	/* Until we're done (re)reading config */
     
     if(!one_process)
+    {
 #ifndef NO_KILLPG
-      killpg(pgrp,SIGHUP);	/* Kill 'em off */
+      if (killpg(pgrp,SIGHUP) < 0)    /* Kill 'em off */
 #else
-      kill(-pgrp,SIGHUP);
+      if (kill(-pgrp,SIGHUP) < 0)
 #endif
+        log_unixerr ("killpg SIGHUP", NULL, NULL, server_conf);
+    }
     
-    if (sd != -1) {
+    if (sd != -1 || listenmaxfd != -1) {
 	reclaim_child_processes(); /* Not when just starting up */
 	log_error ("SIGHUP received.  Attempting to restart", server_conf);
     }
     
+    restart_time = time(NULL);
     clear_pool (pconf);
     ptrans = make_sub_pool (pconf);
     
@@ -913,42 +1403,29 @@ void standalone_main(int argc, char **argv)
     
     default_server_hostnames (server_conf);
 
-    if ((sd = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP)) == -1) {
-        fprintf(stderr,"httpd: could not get socket\n");
-        perror("socket");
-        exit(1);
-    }
+    if (listeners == NULL)
+    {
+	memset((char *) &sa_server, 0, sizeof(sa_server));
+	sa_server.sin_family=AF_INET;
+	sa_server.sin_addr=bind_address;
+	sa_server.sin_port=htons(server_conf->port);
 
-    note_cleanups_for_fd (pconf, sd); /* arrange to close on exec or restart */
-    
-    if((setsockopt(sd,SOL_SOCKET,SO_REUSEADDR,(char *)&one,sizeof(one)))
-       == -1) {
-        fprintf(stderr,"httpd: could not set socket option\n");
-        perror("setsockopt");
-        exit(1);
-    }
-    if((setsockopt(sd,SOL_SOCKET,SO_KEEPALIVE,(void *)&keepalive_value,
-        sizeof(keepalive_value))) == -1) {
-        fprintf(stderr,"httpd: could not set socket option SO_KEEPALIVE\n"); 
-        perror("setsockopt"); 
-        exit(1); 
-    }
+	sd = make_sock(pconf, &sa_server);
+    } else
+    {
+	listen_rec *lr;
+	int fd;
 
-    memset((char *) &sa_server, 0, sizeof(sa_server));
-    sa_server.sin_family=AF_INET;
-    sa_server.sin_addr=bind_address;
-    sa_server.sin_port=htons(server_conf->port);
-    if(bind(sd,(struct sockaddr *) &sa_server,sizeof(sa_server)) == -1) {
-	if (bind_address.s_addr != htonl(INADDR_ANY))
-	    fprintf(stderr,"httpd: could not bind to address %s port %d\n",
-		    inet_ntoa(bind_address), server_conf->port);
-	else
-	    fprintf(stderr,"httpd: could not bind to port %d\n",
-		    server_conf->port);
-        perror("bind");
-        exit(1);
+	listenmaxfd = -1;
+	FD_ZERO(&listenfds);
+	for (lr=listeners; lr != NULL; lr=lr->next)
+	{
+	    fd = make_sock(pconf, &lr->local_addr);
+	    FD_SET(fd, &listenfds);
+	    if (fd > listenmaxfd) listenmaxfd = fd;
+	}
+	sd = -1;
     }
-    listen(sd, 512);
 
     set_signals();
     log_pid(pconf, pid_fname);
@@ -972,17 +1449,19 @@ void standalone_main(int argc, char **argv)
 	    /* Child died... note that it's gone in the scoreboard. */
 	    sync_scoreboard_image();
 	    child_slot = find_child_by_pid (pid);
-	    if (child_slot >= 0) update_child_status (child_slot, SERVER_DEAD);
+	    if (child_slot >= 0)
+		(void)update_child_status (child_slot, SERVER_DEAD,
+		 (request_rec*)NULL);
         }
 
 	sync_scoreboard_image();
 	if ((count_idle_servers() < daemons_min_free)
-	    && (child_slot = find_free_child_num()) >= 0
-	    && child_slot <= daemons_limit)
-	    {
-	    update_child_status(child_slot,SERVER_STARTING);
+	 && (child_slot = find_free_child_num()) >= 0
+	 && child_slot <= daemons_limit) {
+	    (void)update_child_status(child_slot,SERVER_STARTING,
+	     (request_rec*)NULL);
 	    make_child(server_conf, child_slot);
-	    }
+	}
     }
 
 } /* standalone_main */
@@ -995,6 +1474,20 @@ main(int argc, char *argv[])
 {
     int c;
 
+#ifdef AUX
+    (void)set42sig();
+#endif
+
+#ifdef SecureWare
+    if(set_auth_parameters(argc,argv) < 0)
+    	perror("set_auth_parameters");
+    if(getluid() < 0)
+	if(setluid(getuid()) < 0)
+	    perror("setluid");
+    if(setreuid(0, 0) < 0)
+	perror("setreuid");
+#endif
+
     init_alloc();
     pconf = permanent_pool;
     ptrans = make_sub_pool(pconf);
@@ -1022,6 +1515,11 @@ main(int argc, char *argv[])
         }
     }
 
+#ifdef __EMX__
+    printf("%s \n",SERVER_VERSION);
+    printf("OS/2 port by Garey Smiley <garey@slink.com> \n");
+#endif
+
     setup_prelinked_modules();
     
     server_conf = read_config (pconf, ptrans, server_confname);
@@ -1033,26 +1531,52 @@ main(int argc, char *argv[])
     else {
         conn_rec *conn;
 	request_rec *r;
-	struct sockaddr sa_server;
+	struct sockaddr sa_server, sa_client;
+	BUFF *cio;
       
 	open_logs(server_conf, pconf);
 	set_group_privs();
 	default_server_hostnames (server_conf);
 
-        user_id = getuid();
-        group_id = getgid();
+      /* Only try to switch if we're running as root */
+      if(!geteuid() && setuid(user_id) == -1) {
+          log_unixerr("setuid", NULL, "unable to change uid", server_conf);
+          exit (1);
+      }
+
+	c = sizeof(sa_client);
+	if ((getpeername(fileno(stdin), &sa_client, &c)) < 0)
+	{
+/* get peername will fail if the input isn't a socket */
+	    perror("getpeername");
+	    memset(&sa_client, '\0', sizeof(sa_client));
+	}
 
 	c = sizeof(sa_server);
-	if(getsockname(csd, &sa_server, &c) < 0) {
+	if(getsockname(fileno(stdin), &sa_server, &c) < 0) {
 	    perror("getsockname");
 	    fprintf(stderr, "Error getting local address\n");
 	    exit(1);
 	}
 	server_conf->port =ntohs(((struct sockaddr_in *)&sa_server)->sin_port);
-	conn = new_connection (ptrans, server_conf, stdin, stdout,
-			       (struct sockaddr_in *)&sa_server);
+	cio = bcreate(ptrans, B_RDWR);
+	cio->fd = fileno(stdout);
+	cio->fd_in = fileno(stdin);
+	conn = new_connection (ptrans, server_conf, cio,
+			       (struct sockaddr_in *)&sa_client,
+			       (struct sockaddr_in *)&sa_server,-1);
 	r = read_request (conn);
 	if (r) process_request (r); /* else premature EOF (ignore) */
+
+        while (r && conn->keepalive) {
+	  bflush(cio);
+	  destroy_pool(r->pool);
+          r = read_request (conn);
+          if (r) process_request (r);
+        }
+
+	bflush(cio);
+	bclose(cio);
     }
     exit (0);
 }
