@@ -59,7 +59,7 @@
 
 /* Read a request and fill in the fields. */
 
-request_rec *read_request (conn_rec *c);     
+request_rec *read_request (conn_rec *c);
 
 /* Send header for http response */
 
@@ -82,6 +82,7 @@ void send_error_response (request_rec *r, int recursive_error);
  */
 
 int set_content_length (request_rec *r, long length);
+int set_keepalive (request_rec *r);
 int set_last_modified (request_rec *r, time_t mtime);
 
 void add_env_var (array_header *env, char *header_name, char *val);
@@ -103,8 +104,13 @@ long send_fd(FILE *f, request_rec *r);
  * definitions of the macros would get a whole lot hairier.
  */
      
+#if 0
 long rprintf (request_rec *r, char *s, ...);     
+#endif
 int rputc (int c, request_rec *r);     
+int rputs(const char *str, request_rec *r);
+int rvputs(request_rec *r, ...);
+int rprintf(request_rec *r,const char *fmt,...);
      
 /*
  * Index used in custom_responses array for a specific error code
@@ -142,10 +148,16 @@ void client_to_stdout (conn_rec *c);
  *
  * note_basic_auth_failure arranges for the right stuff to be scribbled on
  * the HTTP return so that the client knows how to authenticate itself the
- * next time.
+ * next time. As does note_digest_auth_failure for Digest auth.
+ *
+ * note_auth_failure does the same thing, but will call the correct one
+ * based on the authentication type in use.
+ *
  */
 
+void note_auth_failure(request_rec *r);
 void note_basic_auth_failure(request_rec *r);
+void note_digest_auth_failure(request_rec *r);
 int get_basic_auth_pw (request_rec *r, char **pw);
 
 /*
