@@ -1,64 +1,57 @@
 
-/*-
- * Copyright (c) 1995 The Apache Group. All rights reserved.
- * 
+/* ====================================================================
+ * Copyright (c) 1995 The Apache Group.  All rights reserved.
  *
- * Apache httpd license
- * ====================
- * 
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
  *
- * This is the license for the Apache Server. It covers all the
- * files which come in this distribution, and should never be removed.
- * 
- * The "Apache Group" has based this server, called "Apache", on
- * public domain code distributed under the name "NCSA httpd 1.3".
- * 
- * NCSA httpd 1.3 was placed in the public domain by the National Center 
- * for Supercomputing Applications at the University of Illinois 
- * at Urbana-Champaign.
- * 
- * As requested by NCSA we acknowledge,
- * 
- *  "Portions developed at the National Center for Supercomputing
- *   Applications at the University of Illinois at Urbana-Champaign."
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer. 
  *
- * Copyright on the sections of code added by the "Apache Group" belong
- * to the "Apache Group" and/or the original authors. The "Apache Group" and
- * authors hereby grant permission for their code, along with the
- * public domain NCSA code, to be distributed under the "Apache" name.
- * 
- * Reuse of "Apache Group" code outside of the Apache distribution should
- * be acknowledged with the following quoted text, to be included with any new
- * work;
- * 
- * "Portions developed by the "Apache Group", taken with permission 
- *  from the Apache Server   http://www.apache.org/apache/   "
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in
+ *    the documentation and/or other materials provided with the
+ *    distribution.
  *
+ * 3. All advertising materials mentioning features or use of this
+ *    software must display the following acknowledgment:
+ *    "This product includes software developed by the Apache Group
+ *    for use in the Apache HTTP server project (http://www.apache.org/)."
+ *
+ * 4. The names "Apache Server" and "Apache Group" must not be used to
+ *    endorse or promote products derived from this software without
+ *    prior written permission.
+ *
+ * 5. Redistributions of any form whatsoever must retain the following
+ *    acknowledgment:
+ *    "This product includes software developed by the Apache Group
+ *    for use in the Apache HTTP server project (http://www.apache.org/)."
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
+ * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
+ * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
+ * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
+ * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
+ * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+ * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
+ * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
+ * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
+ * OF THE POSSIBILITY OF SUCH DAMAGE.
+ * ====================================================================
+ *
+ * This software consists of voluntary contributions made by many
+ * individuals on behalf of the Apache Group and was originally based
+ * on public domain software written at the National Center for
+ * Supercomputing Applications, University of Illinois, Urbana-Champaign.
+ * For more information on the Apache Group and the Apache HTTP server
+ * project, please see <http://www.apache.org/>.
  *
- * Permission is hereby granted to anyone to redistribute Apache under
- * the "Apache" name. We do not grant permission for the resale of Apache, but
- * we do grant permission for vendors to bundle Apache free with other software,
- * or to charge a reasonable price for redistribution, provided it is made
- * clear that Apache is free. Permission is also granted for vendors to 
- * sell support for Apache. We explicitly forbid the redistribution of 
- * Apache under any other name.
- * 
- * THIS SOFTWARE IS PROVIDED BY THE CONTRIBUTORS ``AS IS'' AND
- * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
- * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
- * ARE DISCLAIMED.  IN NO EVENT SHALL THE CONTRIBUTORS BE LIABLE
- * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
- * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
- * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
- * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
- * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
- * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
- * SUCH DAMAGE.
- * 
  */
 
 
-
 /*
  * http_request.c: functions to get and process requests
  * 
@@ -77,6 +70,7 @@
 #include "http_protocol.h"
 #include "http_log.h"
 #include "http_main.h"
+#include "scoreboard.h"
 
 /*****************************************************************
  *
@@ -99,6 +93,11 @@ int check_symlinks (char *d, int opts)
     char *lastp;
     int res;
   
+#ifdef __EMX__
+    /* OS/2 dosen't have symlinks */
+    return OK;
+#else
+  
     if (opts & OPT_SYM_LINKS) return OK;
 
     /* Strip trailing '/', if any, off what we're checking; trailing
@@ -133,6 +132,8 @@ int check_symlinks (char *d, int opts)
     if (stat (d, &fi) < 0) return FORBIDDEN;
     
     return (fi.st_uid == lfi.st_uid) ? OK : FORBIDDEN;
+
+#endif    
 }
     
 /* Dealing with the file system to get PATH_INFO
@@ -192,15 +193,29 @@ int directory_walk (request_rec *r)
     core_server_config *sconf = get_module_config (r->server->module_config,
 						   &core_module);
     array_header *sec_array = copy_array (r->pool, sconf->sec);
+    void *per_dir_defaults = r->server->lookup_defaults;
     
     core_dir_config **sec = (core_dir_config **)sec_array->elts;
     int num_sec = sec_array->nelts;
-    void *per_dir_defaults = r->server->lookup_defaults;
     char *test_filename = pstrdup (r->pool, r->filename);
 
     int num_dirs, res;
     int i;
 
+    /* Are we dealing with a file? If not, we can (hopefuly) safely assume
+     * we have a handler that doesn't require one, but for safety's sake,
+     * and so we have something find_types() can get something out of,
+     * fake one. But don't run through the directory entries.
+     */
+
+    if (test_filename == NULL) {
+        r->filename = pstrdup(r->pool, r->uri);
+	r->finfo.st_mode = 0;	/* Not really a file... */
+        r->per_dir_config = per_dir_defaults;
+
+        return OK;
+    }
+
     /* Go down the directory hierarchy.  Where we have to check for symlinks,
      * do so.  Where a .htaccess file has permission to override anything,
      * try to find one.  If either of these things fails, we could poke
@@ -209,6 +224,41 @@ int directory_walk (request_rec *r)
      * for the moment, that's not worth the trouble.
      */
 
+    if (test_filename[0] != '/')
+    {
+/* fake filenames only match Directory sections */
+        void *this_conf, *entry_config;
+        core_dir_config *entry_core;
+	char *entry_dir;
+	int j;
+
+	for (j = 0; j < num_sec; ++j) {
+
+	    entry_config = sec[j];
+	    if (!entry_config) continue;
+	    
+	    entry_core =(core_dir_config *)
+		get_module_config(entry_config, &core_module);
+	    entry_dir = entry_core->d;
+
+	    this_conf = NULL;
+	    if (is_matchexp(entry_dir)) {
+		if (!strcmp_match(test_filename, entry_dir))
+		    this_conf = entry_config;
+	    }
+	    else if (!strncmp (test_filename, entry_dir, strlen(entry_dir)))
+	        this_conf = entry_config;
+
+	    if (this_conf)
+		per_dir_defaults = merge_per_dir_configs (r->pool,
+					   per_dir_defaults, this_conf);
+	}
+
+	r->per_dir_config = per_dir_defaults;
+
+	return OK;
+    }
+
     no2slash (test_filename);
     num_dirs = count_dirs(test_filename);
     get_path_info (r);
@@ -301,6 +351,57 @@ int directory_walk (request_rec *r)
 				 */
 }
 
+int location_walk (request_rec *r)
+{
+    core_server_config *sconf = get_module_config (r->server->module_config,
+						   &core_module);
+    array_header *url_array = copy_array (r->pool, sconf->sec_url);
+    void *per_dir_defaults = r->per_dir_config;
+    
+    core_dir_config **url = (core_dir_config **)url_array->elts;
+    int num_url = url_array->nelts;
+    char *test_location = pstrdup (r->pool, r->uri);
+
+    /* Go through the location entries, and check for matches. */
+
+    if (num_url) {
+        void *this_conf, *entry_config;
+	core_dir_config *entry_core;
+	char *entry_url;
+	int j;
+
+/* 
+ * we apply the directive sections in some order; should really try them
+ * with the most general first.
+ */
+	for (j = 0; j < num_url; ++j) {
+
+	    entry_config = url[j];
+	    if (!entry_config) continue;
+	    
+	    entry_core =(core_dir_config *)
+		get_module_config(entry_config, &core_module);
+	    entry_url = entry_core->d;
+
+	    this_conf = NULL;
+	    if (is_matchexp(entry_url)) {
+		if (!strcmp_match(test_location, entry_url))
+		    this_conf = entry_config;
+	    }
+	    else if (!strncmp (test_location, entry_url, strlen(entry_url)))
+	        this_conf = entry_config;
+
+	    if (this_conf)
+	        per_dir_defaults = merge_per_dir_configs (r->pool,
+					    per_dir_defaults, this_conf);
+	}
+
+	r->per_dir_config = per_dir_defaults;
+    }
+
+    return OK;
+}
+
 /*****************************************************************
  *
  * The sub_request mechanism.
@@ -363,6 +464,9 @@ request_rec *sub_req_lookup_simple (char *new_file, request_rec *r)
     return rnew;
 }
 
+
+static int some_auth_required (request_rec *r);
+
 request_rec *sub_req_lookup_uri (char *new_file, request_rec *r)
 {
     request_rec *rnew;
@@ -373,6 +477,7 @@ request_rec *sub_req_lookup_uri (char *new_file, request_rec *r)
     rnew->connection = r->connection; 
     rnew->server = r->server;
     rnew->request_config = create_request_config (rnew->pool);
+    rnew->htaccess = r->htaccess; /* copy htaccess cache */
     set_sub_req_protocol (rnew, r);
 	
     if (new_file[0] == '/')
@@ -384,7 +489,13 @@ request_rec *sub_req_lookup_uri (char *new_file, request_rec *r)
 	parse_uri (rnew, make_full_path (rnew->pool, udir, new_file));
     }
 	
-    unescape_url (rnew->uri);
+    res = unescape_url (rnew->uri);
+    if (res)
+    {
+	rnew->status = res;
+	return rnew;
+    }
+
     getparents (rnew->uri);
 	
     res = translate_name(rnew);
@@ -398,13 +509,11 @@ request_rec *sub_req_lookup_uri (char *new_file, request_rec *r)
      * However, we'd need to test that the old and new filenames contain the
      * same directory components, so it would require duplicating the start
      * of translate_name.
-     * Maybe it would be easier to implement a cache for directory and
-     * .htaccess stats, or perhaps pass the old request to translate_name()
-     * for it to do the optimisation.
+     * Instead we rely on the cache of .htaccess results.
      */
     
     if ((res = directory_walk (rnew))
-	|| (!auth_type (rnew) ? 0 :
+	|| (!some_auth_required (rnew) ? 0 :
 	     ((res = check_user_id (rnew)) || (res = check_auth (rnew))))
 	|| (res = check_access (rnew))
 	|| (res = find_types (rnew))
@@ -437,6 +546,7 @@ request_rec *sub_req_lookup_file (char *new_file, request_rec *r)
     rnew->connection = r->connection; /* For now... */
     rnew->server = r->server;
     rnew->request_config = create_request_config (rnew->pool);
+    rnew->htaccess = r->htaccess; /* copy htaccess cache */
     set_sub_req_protocol (rnew, r);
 	
     rnew->uri = "INTERNALLY GENERATED file-relative req";
@@ -446,7 +556,7 @@ request_rec *sub_req_lookup_file (char *new_file, request_rec *r)
 	
     if ((res = directory_walk (rnew))
 	|| (res = check_access (rnew))
-	|| (!auth_type (rnew) ? 0 :
+	|| (!some_auth_required (rnew) ? 0 :
 	     ((res = check_user_id (rnew)) && (res = check_auth (rnew))))
 	|| (res = find_types (rnew))
 	|| (res = run_fixups (rnew))
@@ -519,6 +629,11 @@ void die(int type, request_rec *r)
 	    r->no_cache = 1;	/* Do NOT send USE_LOCAL_COPY for
 				 * error documents!
 				 */
+	    /* This redirect needs to be a GET no matter what the original
+	     * method was.
+	     */
+	    r->method = pstrdup(r->pool, "GET");
+	    r->method_number = M_GET;
 	    internal_redirect (custom_response, r);
 	    return;
 	} else {
@@ -547,6 +662,25 @@ static void decl_die (int status, char *phase, request_rec *r)
     else die (status, r);
 }
 
+static int some_auth_required (request_rec *r)
+{
+    /* Is there a require line configured for the type of *this* req? */
+    
+    array_header *reqs_arr = requires (r);
+    require_line *reqs;
+    int i;
+    
+    if (!reqs_arr) return 0;
+    
+    reqs = (require_line *)reqs_arr->elts;
+
+    for (i = 0; i < reqs_arr->nelts; ++i)
+	if (reqs[i].method_mask & (1 << r->method_number))
+	    return 1;
+
+    return 0;
+}
+
 void process_request_internal (request_rec *r)
 {
     int access_status;
@@ -565,9 +699,29 @@ void process_request_internal (request_rec *r)
 	die (BAD_REQUEST, r);
 	return;
     }
-    
-    unescape_url(r->uri);
-    getparents(r->uri);		/* OK --- shrinking transformations... */
+
+    if (!r->hostname && (r->proto_num >= 1001)) {
+        /* Client sent us a HTTP/1.1 or later request without telling
+	 * us the hostname, either with a full URL or a Host: header.
+	 * We therefore need to (as per the 1.1 spec) send an error
+	 */
+        log_reason ("client sent HTTP/1.1 request without hostname",
+		    r->uri, r);
+	die (BAD_REQUEST, r);
+	return;
+    }
+
+    if (!r->proxyreq)
+    {
+	access_status = unescape_url(r->uri);
+	if (access_status)
+	{
+	    die(access_status, r);
+	    return;
+	}
+
+	getparents(r->uri);	/* OK --- shrinking transformations... */
+    }
 
     if ((access_status = translate_name (r))) {
         decl_die (access_status, "translate", r);
@@ -579,12 +733,17 @@ void process_request_internal (request_rec *r)
 	return;
     }	
     
+    if ((access_status = location_walk (r))) {
+        die (access_status, r);
+	return;
+    }	
+    
     if ((access_status = check_access (r)) != 0) {
         decl_die (access_status, "check access", r);
 	return;
     }
     
-    if (auth_type (r)) {
+    if (some_auth_required (r)) {
         if ((access_status = check_user_id (r)) != 0) {
 	    decl_die (access_status, "check user.  No user file?", r);
 	    return;
@@ -612,8 +771,18 @@ void process_request_internal (request_rec *r)
 
 void process_request (request_rec *r)
 {
+#ifdef STATUS
+    int old_stat;
+#endif /* STATUS */
     process_request_internal (r);
+#ifdef STATUS
+    old_stat = update_child_status (r->connection->child_num, SERVER_BUSY_LOG,
+     r);
+#endif /* STATUS */
     log_transaction (r);
+#ifdef STATUS
+    (void)update_child_status (r->connection->child_num, old_stat, r);
+#endif /* STATUS */
 }
 
 table *rename_original_env (pool *p, table *t)
@@ -632,7 +801,7 @@ table *rename_original_env (pool *p, table *t)
     return new;
 }
 
-void internal_redirect (char *new_uri, request_rec *r)
+request_rec *internal_internal_redirect (char *new_uri, request_rec *r)
 {
     request_rec *new = (request_rec *)pcalloc(r->pool, sizeof(request_rec));
     char t[10];			/* Long enough... */
@@ -653,14 +822,10 @@ void internal_redirect (char *new_uri, request_rec *r)
     new->prev = r;
     r->next = new;
     
-    /* We are redirecting.  Treat the internally generated transaction
-     * as a GET, since there is not a chance of its getting POST-style
-     * arguments.   
-     */
-    new->method = "GET";
-    new->method_number = M_GET;
-
     /* Inherit the rest of the protocol info... */
+
+    new->method = r->method;
+    new->method_number = r->method_number;
     
     new->status = r->status;
     new->assbackwards = r->assbackwards;
@@ -673,6 +838,7 @@ void internal_redirect (char *new_uri, request_rec *r)
     new->err_headers_out = r->err_headers_out;
     new->subprocess_env = rename_original_env (r->pool, r->subprocess_env);
     new->notes = make_table (r->pool, 5);
+    new->htaccess = r->htaccess; /* copy .htaccess cache */
     
     new->no_cache = r->no_cache; /* If we've already made up our minds
 				  * about this, don't change 'em back!
@@ -681,6 +847,24 @@ void internal_redirect (char *new_uri, request_rec *r)
     sprintf (t, "%d", r->status);
     table_set (new->subprocess_env, "REDIRECT_STATUS", pstrdup (r->pool, t));
 
+    return new;
+}
+
+void internal_redirect (char *new_uri, request_rec *r)
+{
+    request_rec *new = internal_internal_redirect(new_uri, r);
     process_request_internal (new);
 }
 
+/* This function is designed for things like actions or CGI scripts, when
+ * using AddHandler, and you want to preserve the content type across
+ * an internal redirect.
+ */
+
+void internal_redirect_handler (char *new_uri, request_rec *r)
+{
+    request_rec *new = internal_internal_redirect(new_uri, r);
+    if (r->handler)
+        new->content_type = r->content_type;
+    process_request_internal (new);
+}
