@@ -60,16 +60,27 @@
 
 #include "conf.h"
 #include "alloc.h"
+#include "buff.h"
 
 /* ----------------------------- config dir ------------------------------ */
 
 /* Define this to be the default server home dir. Anything later in this
  * file with a relative pathname will have this added.
  */
-#define HTTPD_ROOT "/etc/httpd"
+#ifdef __EMX__
+/* Set default for OS/2 file system */ 
+#define HTTPD_ROOT "/os2httpd"
+#else
+#define HTTPD_ROOT "/usr/local/etc/httpd"
+#endif
 
 /* Root of server */
-#define DOCUMENT_LOCATION "/var/web/webspace"
+#ifdef __EMX__
+/* Set default for OS/2 file system */ 
+#define DOCUMENT_LOCATION "/os2httpd/docs"
+#else
+#define DOCUMENT_LOCATION "/usr/local/etc/httpd/htdocs"
+#endif
 
 /* Max. number of dynamically loaded modules */
 #define DYNAMIC_MODULE_LIMIT 64
@@ -97,8 +108,18 @@
 #define DEFAULT_GROUP "#-1"
 
 /* The name of the log files */
+#ifdef __EMX__
+/* Set default for OS/2 file system */ 
+#define DEFAULT_XFERLOG "logs/access.log"
+#else
 #define DEFAULT_XFERLOG "logs/access_log"
+#endif
+#ifdef __EMX__
+/* Set default for OS/2 file system */ 
+#define DEFAULT_ERRORLOG "logs/error.log"
+#else
 #define DEFAULT_ERRORLOG "logs/error_log"
+#endif
 #define DEFAULT_PIDLOG "logs/httpd.pid"
 #define DEFAULT_SCOREBOARD "logs/apache_runtime_status"
 
@@ -113,22 +134,27 @@
 #define DEFAULT_TYPE "text/html"
 
 /* Define this to be what your per-directory security files are called */
+#ifdef __EMX__
+/* Set default for OS/2 file system */ 
+#define DEFAULT_ACCESS_FNAME "htaccess"
+#else
 #define DEFAULT_ACCESS_FNAME ".htaccess"
+#endif
 
 /* The name of the server config file */
-#define SERVER_CONFIG_FILE "/etc/httpd/httpd.conf"
+#define SERVER_CONFIG_FILE "conf/httpd.conf"
 
 /* The name of the document config file */
-#define RESOURCE_CONFIG_FILE "/etc/httpd/srm.conf"
+#define RESOURCE_CONFIG_FILE "conf/srm.conf"
 
 /* The name of the MIME types file */
-#define TYPES_CONFIG_FILE "/etc/httpd/mime.types"
+#define TYPES_CONFIG_FILE "conf/mime.types"
 
 /* The name of the access file */
-#define ACCESS_CONFIG_FILE "/etc/httpd/access.conf"
+#define ACCESS_CONFIG_FILE "conf/access.conf"
 
-/* Whether we should enable rfc931 identity checking */
-#define DEFAULT_RFC931 0
+/* Whether we should enable rfc1413 identity checking */
+#define DEFAULT_RFC1413 0
 /* The default directory in user's home dir */
 #define DEFAULT_USER_DIR "public_html"
 
@@ -136,7 +162,12 @@
 #define DEFAULT_PATH "/bin:/usr/bin:/usr/ucb:/usr/bsd:/usr/local/bin"
 
 /* The path to the Bourne shell, for parsed docs */
+#ifdef __EMX__
+/* Set default for OS/2 file system */ 
+#define SHELL_PATH "CMD.EXE"
+#else
 #define SHELL_PATH "/bin/sh"
+#endif
 
 /* The default string lengths */
 #define MAX_STRING_LEN HUGE_STRING_LEN
@@ -145,15 +176,18 @@
 /* The timeout for waiting for messages */
 #define DEFAULT_TIMEOUT 1200
 
+/* The timeout for waiting for keepalive timeout until next request */
+#define DEFAULT_KEEPALIVE_TIMEOUT 15
+
+/* The number of requests to entertain per connection */
+#define DEFAULT_KEEPALIVE 5
+
 /* The size of the server's internal read-write buffers */
 #define IOBUFSIZE 8192
 
 /* The number of header lines we will accept from a client */
 #define MAX_HEADERS 200
 
-/* RFC 1123 format for date - this is what HTTP/1.0 wants */
-#define HTTP_TIME_FORMAT "%a, %d %b %Y %T GMT"
-
 /* Number of servers to spawn off by default --- also, if fewer than
  * this free when the caretaker checks, it will spawn more.
  */
@@ -172,9 +206,17 @@
 /* Limit on the total --- clients will be locked out if more servers than
  * this are needed.  It is intended solely to keep the server from crashing
  * when things get out of hand.
+ *
+ * We keep a hard maximum number of servers, for two reasons --- first off,
+ * in case something goes seriously wrong, we want to stop the fork bomb
+ * short of actually crashing the machine we're running on by filling some
+ * kernel table (I was originally going to make this 256, but it turns out
+ * that that would actually fill the process table on reasonably configured
+ * machines).  Secondly, it keeps the size of the scoreboard file small
+ * enough that we can read the whole thing without worrying too much about
+ * the overhead.
  */
-
-#define DEFAULT_SERVER_LIMIT 150
+#define HARD_SERVER_LIMIT 150
 
 /* Number of requests to try to handle in a single process.  If <= 0,
  * the children don't die off.  That's the default here, since I'm still
@@ -183,15 +225,29 @@
 
 #define DEFAULT_MAX_REQUESTS_PER_CHILD 0
 
-/* ------------------------------ error types ------------------------------ */
+/* If you have altered Apache and wish to change the SERVER_VERSION define
+ * below, please keep to the HTTP/1.0 specification.  This states that
+ * the identification string should consist of product tokens with an optional
+ * slash and version designator.  Sub-products which form a significant part 
+ * of the application can be listed, separated by whitespace.  The tokens
+ * are listed in order of their significance for identifying the application.
+ *
+ * "Product tokens should be short and to the point -- use of them for 
+ * advertizing or other non-essential information is explicitly forbidden."
+ *
+ * Example: "Apache/1.1b3 MrWidget/0.1-alpha" 
+ */
+
+#define SERVER_VERSION "Apache/1.1.1" /* SEE COMMENTS ABOVE */
 
-#define SERVER_VERSION "Apache/1.0.5"
 #define SERVER_PROTOCOL "HTTP/1.0"
 #define SERVER_SUPPORT "http://www.apache.org/"
 
 #define DECLINED -1		/* Module declines to handle */
 #define OK 0			/* Module has handled this stage. */
 
+/* ------------------------------ error types ------------------------------ */
+
 #define DOCUMENT_FOLLOWS 200
 #define REDIRECT 302
 #define USE_LOCAL_COPY 304
@@ -201,15 +257,17 @@
 #define NOT_FOUND 404
 #define SERVER_ERROR 500
 #define NOT_IMPLEMENTED 501
+#define BAD_GATEWAY 502
 #define SERVICE_UNAVAILABLE 503
 #define RESPONSE_CODES 10
 
-#define METHODS 5
+#define METHODS 6
 #define M_GET 0
 #define M_PUT 1
 #define M_POST 2
 #define M_DELETE 3
-#define M_INVALID 4
+#define M_CONNECT 4
+#define M_INVALID 5
 
 #define CGI_MAGIC_TYPE "application/x-httpd-cgi"
 #define INCLUDES_MAGIC_TYPE "text/x-server-parsed-html"
@@ -217,6 +275,7 @@
 #define MAP_FILE_MAGIC_TYPE "application/x-type-map"
 #define ASIS_MAGIC_TYPE "httpd/send-as-is"
 #define DIR_MAGIC_TYPE "httpd/unix-directory"
+#define STATUS_MAGIC_TYPE "application/x-httpd-status"
 
 /* Just in case your linefeed isn't the one the other end is expecting. */
 #define LF 10
@@ -235,9 +294,23 @@
  * pointer is, in fact, zero.
  */
 
+/* This represents the result of calling htaccess; these are cached for
+ * each request.
+ */
+struct htaccess_result
+{
+    char *dir;              /* the directory to which this applies */
+    int override;           /* the overrides allowed for the .htaccess file */
+    void *htaccess;         /* the configuration directives */
+/* the next one, or NULL if no more; N.B. never change this */
+    const struct htaccess_result *next;
+};
+
+
 typedef struct conn_rec conn_rec;
 typedef struct server_rec server_rec;
 typedef struct request_rec request_rec;
+typedef struct listen_rec listen_rec;
 
 struct request_rec {
 
@@ -262,9 +335,13 @@ struct request_rec {
   
   char *the_request;		/* First line of request, so we can log it */
   int assbackwards;		/* HTTP/0.9, "simple" request */
+  int proxyreq;                 /* A proxy request */
   int header_only;		/* HEAD request, as opposed to GET */
   char *protocol;		/* Protocol, as given to us, or HTTP/0.9 */
-  
+  int proto_num;		/* Number version of protocol; 1.1 = 1001 */
+  char *hostname;		/* Host, as set by full URI or Host: */
+  int hostlen;			/* Length of http://host:port in full URI */
+
   char *status_line;		/* Status line, if set by script */
   int status;			/* In any case */
   
@@ -275,8 +352,8 @@ struct request_rec {
   char *method;			/* GET, HEAD, POST, etc. */
   int method_number;		/* M_GET, M_POST, etc. */
 
-  /* int header_bytes_sent; */
-  int bytes_sent;		/* body --- not headers */
+  int sent_bodyct;		/* byte count in stream is for body */
+  long bytes_sent;		/* body byte count, for easy access */
   
   /* MIME header environments, in and out.  Also, an array containing
    * environment variables to be passed to subprocesses, so people can
@@ -297,6 +374,8 @@ struct request_rec {
   table *notes;
 
   char *content_type;		/* Break these out --- we dispatch on 'em */
+  char *handler;		/* What we *really* dispatch on           */
+
   char *content_encoding;
   char *content_language;
   
@@ -305,8 +384,9 @@ struct request_rec {
   /* What object is being requested (either directly, or via include
    * or content-negotiation mapping).
    */
-  
-  char *uri;
+
+  char *uri;                    /* complete URI for a proxy req, or
+                                   URL path for a non-proxy req */
   char *filename;
   char *path_info;
   char *args;			/* QUERY_ARGS, if any */
@@ -319,7 +399,14 @@ struct request_rec {
   
   void *per_dir_config;		/* Options set in config files, etc. */
   void *request_config;		/* Notes on *this* request */
-  
+
+/*
+ * a linked list of the configuration directives in the .htaccess files
+ * accessed by this request.
+ * N.B. always add to the head of the list, _never_ to the end.
+ * that way, a sub request's list can (temporarily) point to a parent's list
+ */
+  const struct htaccess_result *htaccess;
 };
 
 
@@ -332,25 +419,33 @@ struct conn_rec {
   server_rec *server;
   
   /* Information about the connection itself */
-  
-  FILE *client;			/* Connetion to the guy */
-  FILE *request_in;		/* Connection from the guy */
+
+  int child_num;                /* The number of the child handling conn_rec */
+  BUFF *client;			/* Connetion to the guy */
   int aborted;			/* Are we still talking? */
   
   /* Who is the client? */
   
+  struct sockaddr_in local_addr; /* local address */
+  struct sockaddr_in remote_addr;/* remote address */
   char *remote_ip;		/* Client's IP address */
-  char *remote_host;		/* Client's DNS name, if known */
-  char *remote_name;		/* Host ID --- same as remote_host, if known;
-				 * otherwise same as remote_ip.
-				 */
-  char *remote_logname;		/* Only ever set if doing_rfc931 */
-  
-  char *user;			/* If an authentication check was made,
+  char *remote_host;		/* Client's DNS name, if known.
+                                 * NULL if DNS hasn't been checked,
+                                 * "" if it has and no address was found.
+                                 * N.B. Only access this though
+				 * get_remote_host() */
+  char *remote_logname;		/* Only ever set if doing_rfc931
+                                 * N.B. Only access this through
+				 * get_remote_logname() */
+    char *user;			/* If an authentication check was made,
 				 * this gets set to the user name.  We assume
 				 * that there's only one user per connection(!)
 				 */
   char *auth_type;		/* Ditto. */
+
+  int keepalive;		/* Are we using HTTP Keep-Alive? */
+  int keptalive;		/* Did we use HTTP Keep-Alive? */
+  int keepalives;		/* How many times have we used it? */
 };
 
 /* Per-vhost config... */
@@ -368,6 +463,7 @@ struct server_rec {
   
   char *server_admin;
   char *server_hostname;
+  short port;                    /* for redirects, etc. */
   
   /* Log files --- note that transfer log is now in the modules... */
   
@@ -376,6 +472,7 @@ struct server_rec {
   
   /* Module-specific configuration for server, and defaults... */
 
+  int is_virtual;               /* true if this is the virtual server */
   void *module_config;		/* Config vector containing pointers to
 				 * modules' per-server config structures.
 				 */
@@ -384,17 +481,31 @@ struct server_rec {
 				 */
   /* Transaction handling */
 
-  short port;
-  struct in_addr host_addr;	/* Specific address, if "virtual" server */
+  struct in_addr host_addr;	/* The bound address, for this server */
+  short host_port;              /* The bound port, for this server */
   int timeout;			/* Timeout, in seconds, before we give up */
-  int do_rfc931;		/* See if client is advertising a username? */
-  
+  int keep_alive_timeout;	/* Seconds we'll wait for another request */
+  int keep_alive;		/* Maximum requests per connection */
+
+  char *path;			/* Pathname for ServerPath */
+  int pathlen;			/* Length of path */
+
+  char *names;			/* Wildcarded names for HostAlias servers */
+  char *virthost;		/* The name given in <VirtualHost> */
+};
+
+/* These are more like real hosts than virtual hosts */
+struct listen_rec {
+    listen_rec *next;
+    struct sockaddr_in local_addr; /* local IP address and port */
+/* more stuff here, like which protocol is bound to the port */
 };
 
 /* Prototypes for utilities... util.c.
  */
 
 /* Time */
+extern const char month_snames[12][4];
 
 struct tm *get_gmtoff(long *tz);
 char *get_time();
@@ -404,14 +515,17 @@ char *gm_timestr_822(pool *p, time_t t);
 /* String handling */     
      
 char *getword(pool *p, char **line, char stop);
+char *getword_nulls (pool *p, char **line, char stop);
 char *getword_conf (pool *p, char **line);      
+
+char *get_token (pool *p, char **accept_line, int accept_white);
      
 int is_url(char *u);
 extern int unescape_url(char *url);
 void no2slash(char *name);
 void getparents(char *name);
 char *escape_path_segment(pool *p, const char *s);
-char *os_escape_path(pool *p,const char *path);
+char *os_escape_path(pool *p,const char *path,int partial);
 char *escape_uri (pool *p, char *s);
 extern char *escape_html(pool *p, const char *s);
 char *construct_url (pool *p, char *path, server_rec *s);     
@@ -423,6 +537,7 @@ char *make_full_path(pool *a, char *dir, char *f);
      
 int is_matchexp(char *str);
 int strcmp_match(char *str, char *exp);
+int strcasecmp_match(char *str, char *exp);
 char *uudecode (pool *, char *);
 
 void str_tolower (char *);
@@ -440,9 +555,6 @@ int can_exec(struct stat *);
 void chdir_file(char *file);
      
 char *get_local_host(pool *);
-struct in_addr get_local_addr (int sd);
-unsigned long get_virthost_addr (char *hostname, int wild_allowed);
-void get_remote_host(conn_rec *conn);
-int get_portnum(int sd);
-     
+unsigned long get_virthost_addr (char *hostname, short int *port);
 
+extern time_t restart_time;
