@@ -51,13 +51,27 @@
  *
  */
 
+/*
+ * The @ and # characters in the following comment are used by the Debian
+ * package creation procedure to automatically extract a short description
+ * and a full doc. for the module (the lines between # (with the @ lines
+ * removed) go to the short description, the lines between @ (without any
+ * of the # lines) go to the doc).
+ *
+ */
 
 /*
+ * @
+ * #
  * http_alias.c: Stuff for dealing with directory aliases
  * 
+ * Implements the Alias, ScriptAlias and Redirect directives.
+ * #
+ *
  * Original by Rob McCool, rewritten in succession by David Robinson
  * and rst.
- * 
+ * @
+ *
  */
 
 #include "httpd.h"
@@ -66,7 +80,7 @@
 typedef struct {
     char *real;
     char *fake;
-    char *forced_type;
+    char *handler;
 } alias_entry;
 
 typedef struct {
@@ -74,6 +88,9 @@ typedef struct {
     array_header *redirects;
 } alias_server_conf;
 
+typedef struct {
+    array_header *redirects;
+} alias_dir_conf;
 module alias_module;
 
 void *create_alias_config (pool *p, server_rec *s)
@@ -86,6 +103,13 @@ void *create_alias_config (pool *p, server_rec *s)
     return a;
 }
 
+void *create_alias_dir_config (pool *p, char *d)
+{
+    alias_dir_conf *a =
+      (alias_dir_conf *)pcalloc (p, sizeof(alias_dir_conf));
+    a->redirects = make_array (p, 2, sizeof(alias_entry));
+    return a;
+}
 void *merge_alias_config (pool *p, void *basev, void *overridesv)
 {
     alias_server_conf *a =
@@ -98,6 +122,15 @@ void *merge_alias_config (pool *p, void *basev, void *overridesv)
     return a;
 }
 
+void *merge_alias_dir_config (pool *p, void *basev, void *overridesv)
+{
+    alias_dir_conf *a =
+      (alias_dir_conf *)pcalloc (p, sizeof(alias_dir_conf));
+    alias_dir_conf *base = (alias_dir_conf *)basev,
+      *overrides = (alias_dir_conf *)overridesv;
+    a->redirects = append_arrays (p, overrides->redirects, base->redirects);
+    return a;
+}
 char *add_alias(cmd_parms *cmd, void *dummy, char *f, char *r)
 {
     server_rec *s = cmd->server;
@@ -107,19 +140,26 @@ char *add_alias(cmd_parms *cmd, void *dummy, char *f, char *r)
 
     /* XX r can NOT be relative to DocumentRoot here... compat bug. */
     
-    new->fake = f; new->real = r; new->forced_type = cmd->info;
+    new->fake = f; new->real = r; new->handler = cmd->info;
     return NULL;
 }
 
-char *add_redirect(cmd_parms *cmd, void *dummy, char *f, char *url)
+char *add_redirect(cmd_parms *cmd, alias_dir_conf *dirconf, char *f, char *url)
 {
+    alias_entry *new;
     server_rec *s = cmd->server;
-    alias_server_conf *conf =
+    alias_server_conf *serverconf =
         (alias_server_conf *)get_module_config(s->module_config,&alias_module);
-    alias_entry *new = push_array (conf->redirects);
 
     if (!is_url (url)) return "Redirect to non-URL";
-    
+    if ( cmd->path )
+    {
+        new = push_array (dirconf->redirects);
+    }
+    else
+    {
+        new = push_array (serverconf->redirects);
+    }
     new->fake = f; new->real = url;
     return NULL;
 }
@@ -127,9 +167,9 @@ char *add_redirect(cmd_parms *cmd, void *dummy, char *f, char *url)
 command_rec alias_cmds[] = {
 { "Alias", add_alias, NULL, RSRC_CONF, TAKE2, 
     "a fakename and a realname"},
-{ "ScriptAlias", add_alias, CGI_MAGIC_TYPE, RSRC_CONF, TAKE2, 
+{ "ScriptAlias", add_alias, "cgi-script", RSRC_CONF, TAKE2, 
     "a fakename and a realname"},
-{ "Redirect", add_redirect, NULL, RSRC_CONF, TAKE2, 
+{ "Redirect", add_redirect, NULL, OR_FILEINFO, TAKE2, 
     "a document to be redirected, then the destination URL" },
 { NULL }
 };
@@ -178,13 +218,14 @@ char *try_alias_list (request_rec *r, array_header *aliases, int doesc)
         int l = alias_matches (r->uri, p->fake);
 
         if (l > 0) {
-	    if (p->forced_type)
-		table_set (r->notes, "alias-forced-type", p->forced_type);
-			   
+	    if (p->handler) { /* Set handler, and leave a note for mod_cgi */
+	        r->handler = pstrdup(r->pool, p->handler);
+		table_set (r->notes, "alias-forced-type", p->handler);
+	    }
+
 	    if (doesc) {
 		char *escurl;
-		/* would like to use os_escape_path here, but can't */
-		escurl = escape_uri(r->pool, r->uri + l);
+		escurl = os_escape_path(r->pool, r->uri + l, 1);
 		return pstrcat(r->pool, p->real, escurl, NULL);
 	    } else
 		return pstrcat(r->pool, p->real, r->uri + l, NULL);
@@ -197,19 +238,24 @@ char *try_alias_list (request_rec *r, array_header *aliases, int doesc)
 int translate_alias_redir(request_rec *r)
 {
     void *sconf = r->server->module_config;
-    alias_server_conf *conf =
+    alias_server_conf *serverconf =
         (alias_server_conf *)get_module_config(sconf, &alias_module);
     char *ret;
 
+#ifdef __EMX__
+    /* Add support for OS/2 drive names */
+    if ((r->uri[0] != '/' && r->uri[0] != '\0') && r->uri[1] != ':')
+#else    
     if (r->uri[0] != '/' && r->uri[0] != '\0') 
+#endif    
         return BAD_REQUEST;
 
-    if ((ret = try_alias_list (r, conf->redirects, 1)) != NULL) {
+    if ((ret = try_alias_list (r, serverconf->redirects, 1)) != NULL) {
         table_set (r->headers_out, "Location", ret);
         return REDIRECT;
     }
     
-    if ((ret = try_alias_list (r, conf->aliases, 0)) != NULL) {
+    if ((ret = try_alias_list (r, serverconf->aliases, 0)) != NULL) {
         r->filename = ret;
         return OK;
     }
@@ -217,19 +263,28 @@ int translate_alias_redir(request_rec *r)
     return DECLINED;
 }
 
-int type_forced_alias (request_rec *r)
+int fixup_redir(request_rec *r)
 {
-    char *t = table_get (r->notes, "alias-forced-type");
-    if (!t) return DECLINED;
-    r->content_type = t;
-    return OK;
+    void *dconf = r->per_dir_config;
+    alias_dir_conf *dirconf =
+        (alias_dir_conf *)get_module_config(dconf, &alias_module);
+    char *ret;
+
+    /* It may have changed since last time, so try again */
+
+    if ((ret = try_alias_list (r, dirconf->redirects, 1)) != NULL) {
+        table_set (r->headers_out, "Location", ret);
+        return REDIRECT;
+    }
+
+    return DECLINED;
 }
 
 module alias_module = {
    STANDARD_MODULE_STUFF,
    NULL,			/* initializer */
-   NULL,			/* dir config creater */
-   NULL,			/* dir merger --- default is to override */
+   create_alias_dir_config,	/* dir config creater */
+   merge_alias_dir_config,	/* dir merger --- default is to override */
    create_alias_config,		/* server config */
    merge_alias_config,		/* merge server configs */
    alias_cmds,			/* command table */
@@ -238,7 +293,7 @@ module alias_module = {
    NULL,			/* check_user_id */
    NULL,			/* check auth */
    NULL,			/* check access */
-   type_forced_alias,		/* type_checker */
-   NULL,			/* fixups */
+   NULL,			/* type_checker */
+   fixup_redir,			/* fixups */
    NULL				/* logger */
 };
