@@ -32,7 +32,7 @@
  * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
- * IT'S CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
+ * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
  * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
  * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
  * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
@@ -321,7 +321,7 @@ char *find_item(request_rec *r, array_header *list, int path_only) {
             } else if(!path_only) {
                 if(!content_encoding) {
                     if(p->type == BY_TYPE) {
-                        if(!strcmp_match(content_type,p->apply_to))
+                        if(content_type && !strcmp_match(content_type,p->apply_to))
                             return p->data;
                     }
                 } else {
@@ -419,16 +419,42 @@ int insert_readme(char *name, char *readme_fname, int rule, request_rec *r) {
         if(stat(fn,&finfo) == -1)
             return 0;
         plaintext=1;
-        if(rule) rprintf(r,"<HR>%c",LF);
-        rprintf(r,"<PRE>%c",LF);
+        if(rule) rputs("<HR>\n", r);
+        rputs("<PRE>\n", r);
     }
-    else if(rule) rprintf(r,"<HR>%c",LF);
+    else if (rule) rputs("<HR>\n", r);
     if(!(f = pfopen(r->pool,fn,"r")))
         return 0;
-    send_fd(f, r);
+    if (!plaintext)
+	send_fd(f, r);
+    else
+    {
+	char buf[IOBUFSIZE+1];
+	int i, n, c, ch;
+	while (!feof(f))
+	{
+	    do n = fread(buf, sizeof(char), IOBUFSIZE, f);
+	    while (n == -1 && ferror(f) && errno == EINTR);
+	    if (n == -1 || n == 0) break;
+	    buf[n] = '\0';
+	    c = 0;
+	    while (c < n)
+	    {
+		for (i=c; i < n; i++)
+		    if (buf[i] == '<' || buf[i] == '>' || buf[i] == '&') break;
+		ch = buf[i];
+		buf[i] = '\0';
+		rputs(&buf[c], r);
+		if (ch == '<') rputs("&lt;", r);
+		else if (ch == '>') rputs("&gt;", r);
+		else if (ch == '&') rputs("&amp;", r);
+		c = i + 1;
+	    }
+	}
+    }
     pfclose(r->pool, f);
     if(plaintext)
-        rprintf(r,"</PRE>%c",LF);
+        rputs("</PRE>\n", r);
     return 1;
 }
 
@@ -438,13 +464,13 @@ char *find_title(request_rec *r) {
     FILE *thefile = NULL;
     int x,y,n,p;
 
-    if ((!strcmp(r->content_type,"text/html")) && (!r->content_encoding)) {
+    if (r->content_type && !strcmp(r->content_type,"text/html") && !r->content_encoding) {
         if(!(thefile = pfopen(r->pool, r->filename,"r")))
             return NULL;
         n = fread(titlebuf,sizeof(char),MAX_STRING_LEN - 1,thefile);
         titlebuf[n] = '\0';
         for(x=0,p=0;titlebuf[x];x++) {
-            if(titlebuf[x] == find[p]) {
+            if(toupper(titlebuf[x]) == find[p]) {
                 if(!find[++p]) {
                     if((p = ind(&titlebuf[++x],'<')) != -1)
                         titlebuf[x+p] = '\0';
@@ -462,39 +488,6 @@ char *find_title(request_rec *r) {
     return NULL;
 }
 
-
-#ifdef NOTDEF
-
-/* The only call to this function anyplace is commented out in the base
- * code below.  It could be fixed along the same lines as the shell_escape
- * stuff in util.c, but for now, why bother?
- */
-
-void escape_html(char *fn) {
-    register int x,y;
-    char copy[MAX_STRING_LEN];
-
-    strcpy(copy,fn);
-    for(x=0,y=0;copy[y];x++,y++) {
-        if(copy[y] == '<') {
-            strcpy(&fn[x],"&lt;");
-            x+=3;
-        }
-        else if(copy[y] == '>') {
-            strcpy(&fn[x],"&gt;");
-            x+=3;
-        }
-        else if(copy[y] == '&') {
-            strcpy(&fn[x],"&amp;");
-            x+=4;
-        }
-        else
-            fn[x] = copy[y];
-    }
-    fn[x] = '\0';
-}
-#endif
-
 struct ent *make_dir_entry(char *name, int dir_opts,
 			   dir_config_rec *d, request_rec *r)
 {
@@ -573,7 +566,7 @@ char *terminate_description(dir_config_rec *d, char *desc, int dir_opts) {
 void output_directories(struct ent **ar, int n,
 			dir_config_rec *d, request_rec *r, int dir_opts)
 {
-    int x;
+    int x, len;
     char *name = r->uri;
     char *tp;
     pool *scratch = make_sub_pool (r->pool);
@@ -581,20 +574,21 @@ void output_directories(struct ent **ar, int n,
     if(name[0] == '\0') name = "/";
 
     if(dir_opts & FANCY_INDEXING) {
-        rprintf (r, "<PRE>");
+        rputs("<PRE>", r);
         if((tp = find_default_icon(d,"^^BLANKICON^^")))
-            rprintf(r, "<IMG SRC=\"%s\" ALT=\"     \"> ", tp);
-        rprintf (r, "Name                   ");
+            rvputs(r, "<IMG SRC=\"", escape_html(scratch, tp),
+		   "\" ALT=\"     \"> ", NULL);
+        rputs("Name                   ", r);
         if(!(dir_opts & SUPPRESS_LAST_MOD))
-            rprintf(r, "Last modified     ");
+            rputs("Last modified     ", r);
         if(!(dir_opts & SUPPRESS_SIZE))
-            rprintf(r, "Size  ");
+            rputs("Size  ", r);
         if(!(dir_opts & SUPPRESS_DESC))
-            rprintf(r, "Description");
-        rprintf(r,"%c<HR>%c",LF,LF);
+            rputs("Description", r);
+        rputs("\n<HR>\n", r);
     }
     else {
-        rprintf (r, "<UL>");
+        rputs("<UL>", r);
     }
         
     for(x=0;x<n;x++) {
@@ -606,67 +600,77 @@ void output_directories(struct ent **ar, int n,
             char *t = make_full_path (scratch, name, "../");
             getparents(t);
             if(t[0] == '\0') t = "/";
-	    anchor = pstrcat (scratch, "<A HREF=\"", os_escape_path(scratch, t),
+	    anchor = pstrcat (scratch, "<A HREF=\"",
+			      escape_html(scratch, os_escape_path(scratch, t, 0)),
 			      "\">", NULL);
-	    t2 = "Parent Directory</A>";
+	    t2 = "Parent Directory</A>       ";
         }
         else {
 	    t = ar[x]->name;
-	    t2 = pstrdup(scratch, t);
-            if(strlen(t2) > 21) {
-                t2[21] = '\0';
-		t2 = pstrcat (scratch, t2, "</A>..", NULL);
+	    len = strlen(t);
+            if(len > 23) {
+		t2 = pstrdup(scratch, t);
+		t2[21] = '.';
+		t2[22] = '.';
+                t2[23] = '\0';
+		t2 = escape_html(scratch, t2);
+		t2 = pstrcat(scratch, t2, "</A>", NULL);
             } else 
-            /* screws up formatting, but some need it. should fix. */
-	    /* escape_html(t2); */
-            t2 = pstrcat(scratch, t2, "</A>", NULL);
-	    anchor = pstrcat (scratch, "<A HREF=\"", os_escape_path(scratch, t),
+	    {
+		char buff[24]="                       ";
+		t2 = escape_html(scratch, t);
+		buff[23-len] = '\0';
+		t2 = pstrcat(scratch, t2, "</A>", buff, NULL);
+	    }
+	    anchor = pstrcat (scratch, "<A HREF=\"",
+			      escape_html(scratch, os_escape_path(scratch, t, 0)),
 			      "\">", NULL);
         }
 
         if(dir_opts & FANCY_INDEXING) {
             if(dir_opts & ICONS_ARE_LINKS)
-                rprintf (r,"%s",anchor);
+                rputs(anchor, r);
             if((ar[x]->icon) || d->default_icon) {
-                rprintf(r, "<IMG SRC=\"%s\" ALT=\"[%s]\">",
-			ar[x]->icon ? ar[x]->icon : d->default_icon,
-			ar[x]->alt ? ar[x]->alt : "   ");
+                rvputs(r, "<IMG SRC=\"", 
+		       escape_html(scratch, ar[x]->icon ?
+				   ar[x]->icon : d->default_icon),
+		       "\" ALT=\"[", (ar[x]->alt ? ar[x]->alt : "   "),
+		       "]\">", NULL);
             }
             if(dir_opts & ICONS_ARE_LINKS) 
-                rprintf (r, "</A>");
+                rputs("</A>", r);
 
-            rprintf(r," %s%-27.29s", anchor, t2);
+            rvputs(r," ", anchor, t2, NULL);
             if(!(dir_opts & SUPPRESS_LAST_MOD)) {
                 if(ar[x]->lm != -1) {
 		    char time[MAX_STRING_LEN];
                     struct tm *ts = localtime(&ar[x]->lm);
                     strftime(time,MAX_STRING_LEN,"%d-%b-%y %H:%M  ",ts);
-		    rprintf (r, "%s", time);
+		    rputs(time, r);
                 }
                 else {
-                    rprintf(r, "                 ");
+                    rputs("                 ", r);
                 }
             }
             if(!(dir_opts & SUPPRESS_SIZE)) {
                 send_size(ar[x]->size,r);
-                rprintf (r, "  ");
+                rputs("  ", r);
             }
             if(!(dir_opts & SUPPRESS_DESC)) {
                 if(ar[x]->desc) {
-                    rprintf(r,"%s",
-			    terminate_description(d, ar[x]->desc, dir_opts));
+                    rputs(terminate_description(d, ar[x]->desc, dir_opts), r);
                 }
             }
         }
         else
-            rprintf(r, "<LI> %s %s",anchor,t2);
-        rputc(LF,r);
+            rvputs(r, "<LI> ", anchor," ", t2, NULL);
+        rputc('\n', r);
     }
     if(dir_opts & FANCY_INDEXING) {
-        rprintf(r, "</PRE>");
+        rputs("</PRE>", r);
     }
     else {
-        rprintf(r, "</UL>");
+        rputs("</UL>", r);
     }
 }
 
@@ -679,7 +683,7 @@ int dsortf(struct ent **s1,struct ent **s2)
     
 int index_directory(request_rec *r, dir_config_rec *dir_conf)
 {
-    char *title_name = pstrdup (r->pool, r->uri);
+    char *title_name = escape_html(r->pool, r->uri);
     char *title_endp;
     char *name = r->filename;
     
@@ -710,11 +714,11 @@ int index_directory(request_rec *r, dir_config_rec *dir_conf)
     while (title_endp > title_name && *title_endp == '/')
 	*title_endp-- = '\0';
     
-    rprintf (r,"<HEAD><TITLE>Index of %s</TITLE></HEAD><BODY>%c",
-	     title_name, LF);
+    rvputs(r, "<HEAD><TITLE>Index of ", title_name, "</TITLE></HEAD><BODY>\n",
+	   NULL);
 
     if((!(tmp = find_header(dir_conf,r))) || (!(insert_readme(name,tmp,0,r))))
-        rprintf(r,"<H1>Index of %s</H1>%c",title_name,LF);
+        rvputs(r, "<H1>Index of ", title_name, "</H1>\n", NULL);
 
     /* 
      * Since we don't know how many dir. entries there are, put them into a 
@@ -749,10 +753,10 @@ int index_directory(request_rec *r, dir_config_rec *dir_conf)
          if((tmp = find_readme(dir_conf, r)))
              insert_readme(name,tmp,1,r);
      else {
-         rprintf(r, "</UL>");
+         rputs("</UL>", r);
      }
 
-     rprintf(r, "</BODY>");
+     rputs("</BODY>", r);
      return 0;
 }
 
@@ -765,10 +769,15 @@ int handle_dir (request_rec *r)
     char *names_ptr = d->index_names ? d->index_names : DEFAULT_INDEX;
     int allow_opts = allow_options (r);
 
-    if (r->method_number != M_GET) return NOT_IMPLEMENTED;
-    
     if (r->uri[0] == '\0' || r->uri[strlen(r->uri)-1] != '/') {
-        char* ifile = pstrcat (r->pool, r->uri, "/", NULL);
+	char* ifile;
+	if (r->args != NULL)
+        	ifile = pstrcat (r->pool, escape_uri(r->pool, r->uri),
+			"/", "?", r->args, NULL);
+	else
+        	ifile = pstrcat (r->pool, escape_uri(r->pool, r->uri),
+			 "/", NULL);
+
 	table_set (r->headers_out, "Location",
 		   construct_url(r->pool, ifile, r->server));
 	return REDIRECT;
@@ -787,7 +796,13 @@ int handle_dir (request_rec *r)
 	request_rec *rr = sub_req_lookup_uri (name_ptr, r);
            
 	if (rr->status == 200 && rr->finfo.st_mode != 0) {
-            char *new_uri = escape_uri(r->pool, rr->uri);
+	    char* new_uri = escape_uri(r->pool, rr->uri);
+
+	    if (rr->args != NULL)
+		new_uri = pstrcat(r->pool, new_uri, "?", rr->args, NULL);
+	    else if (r->args != NULL)
+		new_uri = pstrcat(r->pool, new_uri, "?", r->args, NULL);
+	
 	    destroy_sub_req (rr);
 	    internal_redirect (new_uri, r);
 	    return OK;
@@ -796,6 +811,8 @@ int handle_dir (request_rec *r)
         destroy_sub_req (rr);
     }
 
+    if (r->method_number != M_GET) return NOT_IMPLEMENTED;
+    
     /* OK, nothing easy.  Trot out the heavy artillery... */
 
     if (allow_opts & OPT_INDEXES) 
