@@ -51,7 +51,26 @@
  *
  */
 
+/*
+ * The @ and # characters in the following comment are used by the Debian
+ * package creation procedure to automatically extract a short description
+ * and a full doc. for the module (the lines between # (with the @ lines
+ * removed) go to the short description, the lines between @ (without any
+ * of the # lines) go to the doc).
+ *
+ */
 
+/*
+ * @
+ * #
+ * mod_log_agent: an agent logging module.
+ *
+ * Provides the AgentLog directive.
+ * #
+ *
+ * See the Apache documentation for usage.
+ * @
+ */
 
 #include "httpd.h"
 #include "http_config.h"
@@ -59,7 +78,12 @@
 module agent_log_module;
 
 static int xfer_flags = ( O_WRONLY | O_APPEND | O_CREAT );
+#ifdef __EMX__
+/* OS/2 dosen't support users and groups */
+static mode_t xfer_mode = ( S_IREAD | S_IWRITE );
+#else
 static mode_t xfer_mode = ( S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
+#endif
 
 typedef struct {
     char *fname;
