@@ -51,7 +51,28 @@
  *
  */
 
+/*
+ * The @ and # characters in the following comment are used by the Debian
+ * package creation procedure to automatically extract a short description
+ * and a full doc. for the module (the lines between # (with the @ lines
+ * removed) go to the short description, the lines between @ (without any
+ * of the # lines) go to the doc).
+ *
+ */
 
+/*
+ * @
+ * #
+ * mod_log_referer: a referer logging module.
+ *
+ * Provides the RefererLog and RefererIgnore directives.
+ * #
+ *
+ * See the Apache documentation for usage.
+ * @
+ */
+
+#include "httpd.h"
 
 #include "httpd.h"
 #include "http_config.h"
@@ -59,7 +80,13 @@
 module referer_log_module;
 
 static int xfer_flags = ( O_WRONLY | O_APPEND | O_CREAT );
+
+#ifdef __EMX__
+/* OS/2 lacks support for users and groups */
+static mode_t xfer_mode = ( S_IREAD | S_IWRITE );
+#else
 static mode_t xfer_mode = ( S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH );
+#endif
 
 typedef struct {
     char *fname;
@@ -192,7 +219,7 @@ int referer_log_transaction(request_rec *orig)
 
 	  for(ptrptr = (char **) cls->referer_ignore_list->elts;
 	      ptrptr < ptrptr2;
-	      ptrptr += cls->referer_ignore_list->elt_size) 
+	      ptrptr = (char **)((char *)ptrptr + cls->referer_ignore_list->elt_size)) 
 	    {
 		if(strstr(referer, *ptrptr))
 		  return OK;
