@@ -175,11 +175,12 @@ typedef struct var_rec {
     int is_pseudo_html;		/* text/html, *or* the INCLUDES_MAGIC_TYPEs */
 
     /* Above are all written-once properties of the variant.  The
-     * two fields below are changed during negotiation:
+     * three fields below are changed during negotiation:
      */
     
     float quality;	
     float level_matched;
+    int mime_stars;
 } var_rec;
 
 /* Something to carry around the state of negotiation (and to keep
@@ -209,13 +210,14 @@ void clean_var_rec (var_rec *mime_info)
     mime_info->content_encoding = "";
     mime_info->content_language = "";
 
-    mime_info->is_pseudo_html = 0.0;
+    mime_info->is_pseudo_html = 0;
     mime_info->level = 0.0;
     mime_info->level_matched = 0.0;
     mime_info->qs = 0.0;
     mime_info->quality = 0.0;
     mime_info->bytes = 0;
     mime_info->lang_index = -1;
+    mime_info->mime_stars = 0;
 }
 
 /* Initializing the relevant fields of a variant record from the
@@ -241,52 +243,6 @@ void set_mime_fields (var_rec *var, accept_rec *mime_info)
  * HTTPD header lines and elsewhere.
  */
 
-/* Retrieve a token, spacing over it and returning a pointer to
- * the first non-white byte afterwards.  Note that these tokens
- * are delimited by semis and commas; and can also be delimited
- * by whitespace at the caller's option.
- */
-
-char *get_token (pool *p, char **accept_line, int accept_white)
-{
-    char *ptr = *accept_line;
-    char *tok_start;
-    char *token;
-    int tok_len;
-  
-    /* Find first non-white byte */
-    
-    while (*ptr && isspace(*ptr))
-      ++ptr;
-
-    tok_start = ptr;
-    
-    /* find token end, skipping over quoted strings.
-     * (comments are already gone).
-     */
-    
-    while (*ptr && (accept_white || !isspace(*ptr))
-	   && *ptr != ';' && *ptr != ',')
-    {
-	if (*ptr++ == '"')
-	    while (*ptr)
-	        if (*ptr++ == '"') break;
-    }
-	  
-    tok_len = ptr - tok_start;
-    token = palloc (p, tok_len + 1);
-    strncpy (token, tok_start, tok_len);
-    token[tok_len] = '\0';
-    
-    /* Advance accept_line pointer to the next non-white byte */
-
-    while (*ptr && isspace(*ptr))
-      ++ptr;
-
-    *accept_line = ptr;
-    return token;
-}
-
 /*
  * Get a single mime type entry --- one media type and parameters;
  * enter the values we recognize into the argument accept_rec
@@ -662,14 +618,24 @@ int read_types_multi (negotiation_state *neg)
 	 */
 	
 	sub_req = sub_req_lookup_file (dir_entry->d_name, r);
-	
+
+	/* If it has a handler, we'll pretend it's a CGI script,
+	 * since that's a good indication of the sort of thing it
+	 * might be doing.
+	 */
+	if (sub_req->handler && !sub_req->content_type)
+	  sub_req->content_type = CGI_MAGIC_TYPE;
+
 	if (sub_req->status != 200 || !sub_req->content_type) continue;
 	
 	/* If it's a map file, we use that instead of the map
 	 * we're building...
 	 */
 
-	if (!strcmp (sub_req->content_type, MAP_FILE_MAGIC_TYPE)) {
+	if (((sub_req->content_type) &&
+	     !strcmp (sub_req->content_type, MAP_FILE_MAGIC_TYPE)) || 
+	    ((sub_req->handler) && 
+	    !strcmp (sub_req->handler, "type-map"))) {
 	    closedir(dirp);
 	    
 	    neg->avail_vars->nelts = 0;
@@ -710,6 +676,9 @@ int read_types_multi (negotiation_state *neg)
  *
  * Note also that if we get an exact match on the media type, we update
  * level_matched for use in level_cmp below...
+ * 
+ * We also give a value for mime_stars, which is used later. It should
+ * be 1 for star/star, 2 for type/star and 3 for type/subtype.
  */
 
 int mime_match (accept_rec *accept, var_rec *avail)
@@ -718,16 +687,24 @@ int mime_match (accept_rec *accept, var_rec *avail)
     char *avail_type = avail->type_name;
     int len = strlen(accept_type);
   
-    if (accept_type[0] == '*')	/* Anything matches star/star */
+    if (accept_type[0] == '*')	{ /* Anything matches star/star */
+        if (avail->mime_stars < 1)
+	  avail->mime_stars = 1;
 	return 1; 
-    else if (accept_type[len - 1] == '*')
-	return !strncmp (accept_type, avail_type, len - 2);
+    }
+    else if ((accept_type[len - 1] == '*') &&
+	     !strncmp (accept_type, avail_type, len - 2)) {
+        if (avail->mime_stars < 2)
+	  avail->mime_stars = 2;
+	return 1;
+    }
     else if (!strcmp (accept_type, avail_type)
 	     || (!strcmp (accept_type, "text/html")
 		 && (!strcmp(avail_type, INCLUDES_MAGIC_TYPE)
 		     || !strcmp(avail_type, INCLUDES_MAGIC_TYPE3)))) {
 	if (accept->level >= avail->level) {
 	    avail->level_matched = avail->level;
+	    avail->mime_stars = 3;
 	    return 1;
 	}
     }
@@ -886,7 +863,7 @@ int find_encoding (array_header *accept_encodings, char *enc)
     accept_rec *accs = (accept_rec *)accept_encodings->elts;
     int i;
 
-    if (is_identity_encoding(enc)) return 1.0;
+    if (is_identity_encoding(enc)) return 1;
 
     for (i = 0; i < accept_encodings->nelts; ++i)
 	if (!strcmp (enc, accs[i].type_name))
@@ -906,7 +883,7 @@ void do_encodings (negotiation_state *neg)
 	return;
 
     /* Lose any variant with an unacceptable content encoding */
-
+    
     for (i = 0; i < neg->avail_vars->nelts; ++i)
 	if (var_recs[i].quality > 0
 	    && !find_encoding (neg->accept_encodings,
@@ -996,6 +973,10 @@ var_rec *best_match(negotiation_state *neg)
 	     * tied variants by whatever means it likes.  This server
 	     * breaks ties as follows, in order:
 	     *
+	     * By perferring non-wildcard entries to those with
+	     * wildcards. The spec specifically says we should
+	     * do this, and it makes a lot of sense.
+	     *
 	     * By order of languages in Accept-language, to give the
 	     * client a way to specify a language preference.  I'd prefer
 	     * to give this precedence over media type, but the standard
@@ -1015,15 +996,16 @@ var_rec *best_match(negotiation_state *neg)
 		
 	    if (q > best_quality
 		|| (q == best_quality
-		    && (variant->lang_index < best->lang_index
-			|| (variant->lang_index == best->lang_index
-			    && ((levcmp = level_cmp (variant, best)) == 1
-				|| (levcmp == 0
-				    && !strcmp (variant->type_name,
-						best->type_name)
-				    && (find_content_length(neg, variant)
-					<
-					find_content_length(neg, best))))))))
+		    && ((variant->mime_stars > best->mime_stars)
+			|| (variant->lang_index < best->lang_index
+			    || (variant->lang_index == best->lang_index
+				&& ((levcmp = level_cmp (variant, best)) == 1
+				    || (levcmp == 0
+					&& !strcmp (variant->type_name,
+						    best->type_name)
+					&& (find_content_length(neg, variant)
+					    <
+				find_content_length(neg, best)))))))))
 	    {
 		best = variant;
 		best_quality = q;
@@ -1107,6 +1089,7 @@ int handle_multi (request_rec *r)
     
     if (!do_cache_negotiated_docs(r->server)) r->no_cache = 1;
     r->filename = sub_req->filename;
+    r->handler = sub_req->handler;
     r->content_type = sub_req->content_type;
     r->content_encoding = sub_req->content_encoding;
     r->content_language = sub_req->content_language;
@@ -1117,6 +1100,7 @@ int handle_multi (request_rec *r)
 
 handler_rec negotiation_handlers[] = {
 { MAP_FILE_MAGIC_TYPE, handle_map_file },
+{ "type-map", handle_map_file },
 { NULL }
 };
 
